# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Default Hyperparameter configuration."""
from __future__ import annotations

import dataclasses


@dataclasses.dataclass(unsafe_hash=True)
class MeshRules:
  embed: str | None = None
  mlp: str | None = None
  kv: str | None = None
  vocab: str | None = None

  def __call__(self, *keys: str) -> tuple[str, ...]:
    return tuple(getattr(self, key) for key in keys)


@dataclasses.dataclass(unsafe_hash=True)
class Config:
  # Path to load or store sentencepiece vocab file.
  vocab_path: str | None = None
  # Vocabulary size if `vocab_path` is not given.
  vocab_size: int = 30_000
  # Maximum number of characters to use for training.
  max_corpus_chars: int = 10**7
  # Name of TFDS translation dataset to use.
  dataset_name: str = 'lm1b'
  # Optional name of TFDS translation dataset to use for evaluation.
  eval_dataset_name: str = 'lm1b'
  # Optional name of TFDS split to use for evaluation.
  eval_split: str = 'test'
  # Per device batch size for training.
  per_device_batch_size: int = 32
  # Per device batch size for training.
  eval_per_device_batch_size: int = 32
  # Sampling temperature for language model inference.
  sampling_temperature: float = 0.6
  # Top k cutoff for logit sampling. If 0 then no top-k cutoff is used.
  sampling_top_k: int = 20
  # Number of steps to take during training.
  num_train_steps: int = 500_000
  # Number of steps to take during evaluation.
  # Large enough to evaluate all samples: 306_688 / (32 * 8) = 1198
  num_eval_steps: int = 2_000
  # Number of steps to generate predictions.
  # -1 will use the whole eval dataset.
  num_predict_steps: int = -1
  # Base learning rate.
  learning_rate: float = 0.0016
  # Linear learning rate warmup.
  warmup_steps: int = 1000
  # Cross entropy loss label smoothing.
  label_smoothing: float = 0.0
  # Decay factor for AdamW style weight decay.
  weight_decay: float = 0.1
  # Maximum length cutoff for training examples.
  max_target_length: int = 128
  # Maximum length cutoff for eval examples.
  max_eval_target_length: int = 512
  # Maximum length cutoff for predicted tokens.
  max_predict_length: int = 50
  # Final logit transform uses embedding matrix transpose.
  logits_via_embedding: bool = False
  # Number of transformer layers.
  num_layers: int = 6
  # Size of query/key/value for attention.
  qkv_dim: int = 512
  # Size of embeddings.
  emb_dim: int = 512
  # Size of the MLP.
  mlp_dim: int = 2048
  # Number of attention heads.
  num_heads: int = 8
  # Dropout rate.
  dropout_rate: float = 0.1
  # Attention dropout rate.
  attention_dropout_rate: float = 0.1
  # Whether to save model checkpoints.
  save_checkpoints: bool = True
  # Whether to restore from existing model checkpoints.
  restore_checkpoints: bool = True
  # Save a checkpoint every these number of steps.
  checkpoint_every_steps: int = 10_000
  # Frequency of eval during training, e.g. every 1_000 steps.
  eval_every_steps: int = 1_000
  # Use bfloat16 mixed precision training instead of float32.
  use_bfloat16: bool = True
  # Integer for PRNG random seed.
  seed: int = 0
  # Prompt for language model sampling,
  # taken from MaxText (https://github.com/google/maxtext/blob/main/MaxText/configs/base.yml).
  prompts: str = 'I love to '
  # Parallelism
  mesh_axes: tuple[str, ...] = ('data', 'fsdp', 'tensor')
  axis_rules: MeshRules = MeshRules(
    embed='fsdp',
    mlp='tensor',
    kv='tensor',
    vocab='tensor',
  )
  data_sharding: tuple[str, ...] = ('data',)
  # One axis for each parallelism type may hold a placeholder (-1)
  # value to auto-shard based on available slices and devices.
  # By default, product of the DCN axes should equal number of slices
  # and product of the ICI axes should equal number of devices per slice.
  # ICI (Inter-Chip Interconnection): A high-speed connection between
  # sets of TPU chips, which form the TPU network.
  # DCN (Data Center Network): A connection between the TPU networks;
  # not as fast as ICI.
  # ICI has around 100x the bandwidth of DCN, but it is not a general
  # purpose connection, which is why DCN is necessary for scaling to
  # extremely large ML models.
  dcn_data_parallelism: int = -1
  dcn_fsdp_parallelism: int = 1
  dcn_tensor_parallelism: int = 1
  ici_data_parallelism: int = 1
  ici_fsdp_parallelism: int = -1
  ici_tensor_parallelism: int = 1

  def replace(self, **kwargs):
    return dataclasses.replace(self, **kwargs)


def get_config():
  """Get the default hyperparameter configuration."""
  config = Config()
  return config



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for flax.errors."""

from absl.testing import absltest
from flax.errors import FlaxError, ScopeVariableNotFoundError
import pickle

class ErrorrsTest(absltest.TestCase):
  def test_exception_can_be_pickled(self):
    # tests the new __reduce__ method fixes bug reported in issue #4000
    ex = ScopeVariableNotFoundError('varname', 'collection', 'scope')
    pickled_ex = pickle.dumps(ex)
    unpicked_ex = pickle.loads(pickled_ex)
    self.assertIsInstance(unpicked_ex, FlaxError)
    self.assertIn('varname', str(unpicked_ex))
    self.assertIn('#flax.errors.ScopeVariableNotFoundError', str(unpicked_ex))
    self.assertNotIn('#flax.errors.FlaxError', str(unpicked_ex))


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from absl.testing import absltest

from flax import nnx


class TestFilters(absltest.TestCase):
  def test_path_contains(self):
    class Model(nnx.Module):
      def __init__(self, rngs):
        self.backbone = nnx.Linear(2, 3, rngs=rngs)
        self.head = nnx.Linear(3, 10, rngs=rngs)

    model = Model(nnx.Rngs(0))

    head_state = nnx.state(model, nnx.PathContains('head'))

    self.assertIn('head', head_state)
    self.assertNotIn('backbone', head_state)


# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Agent utilities, incl. choosing the move and running in separate process."""

import collections
import functools
import multiprocessing
from typing import Any, Callable

import flax
import jax
import numpy as np

import env_utils


@functools.partial(jax.jit, static_argnums=0)
def policy_action(
    apply_fn: Callable[..., Any],
    params: flax.core.frozen_dict.FrozenDict,
    state: np.ndarray,
):
  """Forward pass of the network.

  Args:
    params: the parameters of the actor-critic model
    module: the actor-critic model
    state: the input for the forward pass

  Returns:
    out: a tuple (log_probabilities, values)
  """
  out = apply_fn({'params': params}, state)
  return out


ExpTuple = collections.namedtuple(
    'ExpTuple', ['state', 'action', 'reward', 'value', 'log_prob', 'done']
)


class RemoteSimulator:
  """Wrap functionality for an agent emulating Atari in a separate process.

  An object of this class is created for every agent.
  """

  def __init__(self, game: str):
    """Start the remote process and create Pipe() to communicate with it."""
    parent_conn, child_conn = multiprocessing.Pipe()
    self.proc = multiprocessing.Process(
        target=rcv_action_send_exp, args=(child_conn, game)
    )
    self.proc.daemon = True
    self.conn = parent_conn
    self.proc.start()


def rcv_action_send_exp(conn, game: str):
  """Run the remote agents.

  Receive action from the main learner, perform one step of simulation and
  send back collected experience.
  """
  env = env_utils.create_env(game, clip_rewards=True)
  while True:
    obs = env.reset()
    done = False
    # Observations fetched from Atari env need additional batch dimension.
    state = obs[None, ...]
    while not done:
      conn.send(state)
      action = conn.recv()
      obs, reward, done, _ = env.step(action)
      next_state = obs[None, ...] if not done else None
      experience = (state, action, reward, done)
      conn.send(experience)
      if done:
        break
      state = next_state



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Helper functions."""

from __future__ import annotations

from typing import Callable, Optional, Tuple, TypeVar, Union
import flax
from flax import nnx
from flax.typing import VariableDict  # pylint: disable=g-importing-member,g-multiple-import

M = TypeVar('M', bound='nnx.Module')


def _flatten_path(path: Tuple[Union[str, int], ...]) -> str:
  def f(item) -> str:
    if isinstance(item, str):
      return f'{item}'
    elif isinstance(item, int):
      return f'[{item}]'
    else:
      raise ValueError(f'Unexpected type {type(item)}')

  return '.'.join([f(item) for item in path]).replace('.[', '[')


def module_from_linen_variables(
    module_factory: Callable[[], M],
    variables: VariableDict,
    map_key_fn: Optional[
        Callable[[Tuple[str, ...]], Tuple[Union[str, int], ...]]
    ] = None,
) -> M:
  """Returns an `nnx.Module` initialized with the `variables` of a linen module.

  Args:
    module_factory: A no-args callable that returns an `nnx.Module`.
    variables: A dictionary of variables.
    map_key_fn: An optional function for mapping keys in the `variables`
      dictionary to keys in the `nnx.Module`'s state. If not provided it is
      assumed that after removing the collection name the keys in the
      `variables` dictionary are the same as the keys in the `nnx.Module`'s
      state.
  """
  if map_key_fn is None:

    def map_key_fn(path: Tuple[str, ...]) -> Tuple[Union[str, int], ...]:
      return path[1:] if 'params' in variables else path

  mdl: M = nnx.eval_shape(module_factory)
  graph_def, state = nnx.split(mdl)
  state = state.flat_state()
  for path, val in flax.traverse_util.flatten_dict(variables).items():
    mapped_path = map_key_fn(path)
    if mapped_path not in state:
      raise ValueError(
          f"'{mdl.__class__.__name__}.{_flatten_path(mapped_path)}' doesn't "
          f' exist (original path={path}).'
      )
    state[mapped_path].value = val
  state = nnx.State.from_flat_path(state)

  return nnx.merge(graph_def, state)



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for flax.experimental.nnx.traversal."""
from absl.testing import absltest
from flax.core import freeze
from flax.nnx import traversals
import jax

# Parse absl flags test_srcdir and test_tmpdir.
jax.config.parse_flags_with_absl()


class TraversalTest(absltest.TestCase):
  def test_flatten_mapping(self):
    xs = {'foo': 1, 'bar': {'a': 2, 'b': {}}}
    flat_xs = traversals.flatten_mapping(xs)
    self.assertEqual(
      flat_xs,
      {
        ('foo',): 1,
        ('bar', 'a'): 2,
      },
    )
    flat_xs = traversals.flatten_mapping(freeze(xs))
    self.assertEqual(
      flat_xs,
      {
        ('foo',): 1,
        ('bar', 'a'): 2,
      },
    )
    flat_xs = traversals.flatten_mapping(xs, sep='/')
    self.assertEqual(
      flat_xs,
      {
        'foo': 1,
        'bar/a': 2,
      },
    )

  def test_unflatten_mapping(self):
    expected_xs = {'foo': 1, 'bar': {'a': 2}}
    xs = traversals.unflatten_mapping(
      {
        ('foo',): 1,
        ('bar', 'a'): 2,
      }
    )
    self.assertEqual(xs, expected_xs)
    xs = traversals.unflatten_mapping(
      {
        'foo': 1,
        'bar/a': 2,
      },
      sep='/',
    )
    self.assertEqual(xs, expected_xs)

  def test_flatten_mapping_keep_empty(self):
    ys = {'a': {}}
    xs = {'foo': 1, 'bar': {'a': 2, 'b': {}}}
    flat_ys = traversals.flatten_mapping(ys, keep_empty_nodes=True)
    flat_xs = traversals.flatten_mapping(xs, keep_empty_nodes=True)
    empty_node = flat_ys[('a',)]
    self.assertEqual(
      flat_xs,
      {
        ('foo',): 1,
        ('bar', 'a'): 2,
        ('bar', 'b'): empty_node,
      },
    )
    xs_restore = traversals.unflatten_mapping(flat_xs)
    self.assertEqual(xs, xs_restore)

  def test_flatten_mapping_is_leaf(self):
    xs = {'foo': {'c': 4}, 'bar': {'a': 2, 'b': {}}}
    flat_xs = traversals.flatten_mapping(
      xs, is_leaf=lambda k, x: len(k) == 1 and len(x) == 2
    )
    self.assertEqual(
      flat_xs,
      {
        ('foo', 'c'): 4,
        ('bar',): {'a': 2, 'b': {}},
      },
    )
    xs_restore = traversals.unflatten_mapping(flat_xs)
    self.assertEqual(xs, xs_restore)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The Flax Module system."""


# pylint: disable=g-multiple-import,useless-import-alias
# re-export commonly used modules and functions
from flax.core import (
    DenyList as DenyList,
    FrozenDict as FrozenDict,
    broadcast as broadcast,
    meta as meta,
)
from flax.core.meta import (
    PARTITION_NAME as PARTITION_NAME,
    Partitioned as Partitioned,
    get_partition_spec as get_partition_spec,
    get_sharding as get_sharding,
    unbox as unbox,
    with_partitioning as with_partitioning,
)
from .activation import (
    GeGLU as GeGLU,
    PReLU as PReLU,
    celu as celu,
    elu as elu,
    gelu as gelu,
    glu as glu,
    hard_sigmoid as hard_sigmoid,
    hard_silu as hard_silu,
    hard_swish as hard_swish,
    hard_tanh as hard_tanh,
    leaky_relu as leaky_relu,
    log_sigmoid as log_sigmoid,
    log_softmax as log_softmax,
    logsumexp as logsumexp,
    normalize as normalize,
    one_hot as one_hot,
    relu6 as relu6,
    relu as relu,
    selu as selu,
    sigmoid as sigmoid,
    silu as silu,
    soft_sign as soft_sign,
    softmax as softmax,
    softplus as softplus,
    standardize as standardize,
    swish as swish,
    tanh as tanh,
)
from .attention import (
    MultiHeadAttention as MultiHeadAttention,
    MultiHeadDotProductAttention as MultiHeadDotProductAttention,
    SelfAttention as SelfAttention,
    combine_masks as combine_masks,
    dot_product_attention_weights as dot_product_attention_weights,
    dot_product_attention as dot_product_attention,
    make_attention_mask as make_attention_mask,
    make_causal_mask as make_causal_mask,
)
from .batch_apply import BatchApply as BatchApply
from .combinators import Sequential as Sequential
from .fp8_ops import Fp8DotGeneralOp as Fp8DotGeneralOp
from .initializers import (
    ones_init as ones_init,
    ones as ones,
    zeros_init as zeros_init,
    zeros as zeros,
)
from .linear import (
    ConvLocal as ConvLocal,
    ConvTranspose as ConvTranspose,
    Conv as Conv,
    DenseGeneral as DenseGeneral,
    Dense as Dense,
    Einsum as Einsum,
    Embed as Embed,
)
from .module import (
    Module as Module,
    Variable as Variable,
    apply as apply,
    compact as compact,
    compact_name_scope as compact_name_scope,
    disable_named_call as disable_named_call,
    enable_named_call as enable_named_call,
    init_with_output as init_with_output,
    init as init,
    intercept_methods as intercept_methods,
    merge_param as merge_param,
    nowrap as nowrap,
    override_named_call as override_named_call,
)
from .normalization import (
    BatchNorm as BatchNorm,
    GroupNorm as GroupNorm,
    InstanceNorm as InstanceNorm,
    LayerNorm as LayerNorm,
    RMSNorm as RMSNorm,
    SpectralNorm as SpectralNorm,
    WeightNorm as WeightNorm
)
from .pooling import (avg_pool as avg_pool, max_pool as max_pool, pool as pool)
from .recurrent import (
    Bidirectional as Bidirectional,
    ConvLSTMCell as ConvLSTMCell,
    SimpleCell as SimpleCell,
    GRUCell as GRUCell,
    MGUCell as MGUCell,
    LSTMCell as LSTMCell,
    OptimizedLSTMCell as OptimizedLSTMCell,
    RNNCellBase as RNNCellBase,
    RNN as RNN,
)
from .spmd import (
    LogicallyPartitioned as LogicallyPartitioned,
    get_logical_axis_rules as get_logical_axis_rules,
    logical_axis_rules as logical_axis_rules,
    logical_to_mesh,
    logical_to_mesh_axes,
    logical_to_mesh_sharding,
    set_logical_axis_rules as set_logical_axis_rules,
    with_logical_constraint,
    with_logical_partitioning as with_logical_partitioning,
)
from .stochastic import Dropout as Dropout
from .summary import tabulate
from .transforms import (
    add_metadata_axis,
    checkpoint as checkpoint,
    cond as cond,
    custom_vjp as custom_vjp,
    jit as jit,
    jvp as jvp,
    map_variables as map_variables,
    named_call as named_call,
    remat_scan as remat_scan,
    remat as remat,
    scan as scan,
    switch as switch,
    vjp as vjp,
    grad as grad,
    value_and_grad as value_and_grad,
    vmap as vmap,
    while_loop as while_loop,
)
# pylint: enable=g-multiple-import



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from absl.testing import absltest
from absl.testing import parameterized
from flax import nnx
import jax
import jax.numpy as jnp


class TestMetrics(parameterized.TestCase):

  def test_split_merge(self):
    logits = jax.random.normal(jax.random.key(0), (5, 2))
    labels = jnp.array([1, 1, 0, 1, 0])
    logits2 = jax.random.normal(jax.random.key(1), (5, 2))
    labels2 = jnp.array([0, 1, 1, 1, 1])

    accuracy = nnx.metrics.Accuracy()
    accuracy.update(logits=logits, labels=labels)
    graphdef, state = accuracy.split()
    accuracy = nnx.merge(graphdef, state)
    self.assertEqual(accuracy.compute(), 0.6)
    accuracy.update(logits=logits2, labels=labels2)
    self.assertEqual(accuracy.compute(), 0.7)

  def test_welford(self):
    values = jax.random.normal(jax.random.key(0), (5, 2))

    welford = nnx.metrics.Welford()
    welford.update(values=values)
    graphdef, state = welford.split()
    welford = nnx.merge(graphdef, state)
    expected = nnx.metrics.Statistics(
        mean=values.mean(),
        standard_deviation=values.std(),
        standard_error_of_mean=values.std() / jnp.sqrt(values.size),
    )
    computed = welford.compute()
    self.assertAlmostEqual(computed.mean, expected.mean, )
    self.assertAlmostEqual(computed.standard_deviation, expected.standard_deviation)
    self.assertAlmostEqual(
        computed.standard_error_of_mean, expected.standard_error_of_mean
    )

  def test_welford_large(self):
    values = jax.random.normal(jax.random.key(0), (5, 2)) + 1e16

    welford = nnx.metrics.Welford()
    welford.update(values=values)
    graphdef, state = welford.split()
    welford = nnx.merge(graphdef, state)
    expected = nnx.metrics.Statistics(
        mean=values.mean(),
        standard_deviation=values.std(),
        standard_error_of_mean=values.std() / jnp.sqrt(values.size),
    )
    computed = welford.compute()
    self.assertAlmostEqual(computed.mean, expected.mean)
    self.assertAlmostEqual(computed.standard_deviation, expected.standard_deviation)
    self.assertAlmostEqual(
        computed.standard_error_of_mean, expected.standard_error_of_mean
    )

  def test_welford_many(self):
    values = jax.random.normal(jax.random.key(0), (50_000,))

    welford = nnx.metrics.Welford()
    welford.update(values=values)
    graphdef, state = welford.split()
    welford = nnx.merge(graphdef, state)
    computed = welford.compute()
    self.assertAlmostEqual(
        computed.mean, 0.0, delta=3 * computed.standard_error_of_mean
    )
    self.assertAlmostEqual(computed.standard_deviation, 1.0, places=2)

  def test_multimetric(self):
    logits = jax.random.normal(jax.random.key(0), (5, 2))
    labels = jnp.array([1, 1, 0, 1, 0])
    logits2 = jax.random.normal(jax.random.key(1), (5, 2))
    labels2 = jnp.array([0, 1, 1, 1, 1])
    batch_loss = jnp.array([1, 2, 3, 4])
    batch_loss2 = jnp.array([3, 2, 1, 0])

    metrics = nnx.MultiMetric(
        accuracy=nnx.metrics.Accuracy(), loss=nnx.metrics.Average()
    )
    values = metrics.compute()
    self.assertTrue(jnp.isnan(values['accuracy']))
    self.assertTrue(jnp.isnan(values['loss']))

    metrics.update(logits=logits, labels=labels, values=batch_loss)
    values = metrics.compute()
    self.assertEqual(values['accuracy'], 0.6)
    self.assertEqual(values['loss'], 2.5)

    metrics.update(logits=logits2, labels=labels2, values=batch_loss2)
    values = metrics.compute()
    self.assertEqual(values['accuracy'], 0.7)
    self.assertEqual(values['loss'], 2.0)

    metrics.reset()
    values = metrics.compute()
    self.assertTrue(jnp.isnan(values['accuracy']))
    self.assertTrue(jnp.isnan(values['loss']))


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""Parallel BLEU score calculation.

This version of BLEU calculation is derived from the MLPerf transformer reference.
Tries to match SacreBLEU metric reasonably well, but is not identical.

Refs:
    tokenizer at:
    https://github.com/tensorflow/models/blob/master/official/transformer/utils/tokenizer.py
    original preprocessing tokenizer:
    https://github.com/moses-smt/mosesdecoder/blob/master/scripts/generic/mteval-v14.pl#L954-L983
    original t2t code:
    https://github.com/tensorflow/tensor2tensor/blob/master/tensor2tensor/utils/bleu_hook.py

Usage:
    refs = '''food bar brown cow
    blee bloo dog sat
    or please take me out
    '''
    hyps = '''foo bar brown cow
    blee bloo dog sit
    please do take me out
    '''
    bleu_local(refs.split("\n"), hyps.split("\n"))  # 39.65
"""

import collections
import math
import re
import sys
import unicodedata

import numpy as np


class UnicodeRegex:
  """Ad-hoc hack to recognize all punctuation and symbols."""

  def __init__(self):
    punctuation = self.property_chars("P")
    self.nondigit_punct_re = re.compile(r"([^\d])([" + punctuation + r"])")
    self.punct_nondigit_re = re.compile(r"([" + punctuation + r"])([^\d])")
    self.symbol_re = re.compile("([" + self.property_chars("S") + "])")

  def property_chars(self, prefix):
    return "".join(
        chr(x)
        for x in range(sys.maxunicode)
        if unicodedata.category(chr(x)).startswith(prefix)
    )


uregex = UnicodeRegex()


def bleu_tokenize(string):
  r"""Tokenize a string following the official BLEU implementation.

  See https://github.com/moses-smt/mosesdecoder/'
           'blob/master/scripts/generic/mteval-v14.pl#L954-L983
  In our case, the input string is expected to be just one line
  and no HTML entities de-escaping is needed.
  So we just tokenize on punctuation and symbols,
  except when a punctuation is preceded and followed by a digit
  (e.g. a comma/dot as a thousand/decimal separator).

  Note that a number (e.g. a year) followed by a dot at the end of sentence
  is NOT tokenized, i.e. the dot stays with the number because
  `s/(\p{P})(\P{N})/ $1 $2/g` does not match this case (unless we add a
  space after each sentence). However, this error is already in the
  original mteval-v14.pl and we want to be consistent with it.

  Args:
    string: the input string

  Returns:
    a list of tokens
  """
  string = uregex.nondigit_punct_re.sub(r"\1 \2 ", string)
  string = uregex.punct_nondigit_re.sub(r" \1 \2", string)
  string = uregex.symbol_re.sub(r" \1 ", string)
  return string.split()


def _get_ngrams(segment, max_order):
  """Extracts all n-grams up to a given maximum order from an input segment.

  Args:
    segment: text segment from which n-grams will be extracted.
    max_order: maximum length in tokens of the n-grams returned by this methods.

  Returns:
    The Counter containing all n-grams up to max_order in segment
    with a count of how many times each n-gram occurred.
  """
  ngram_counts = collections.Counter()
  for order in range(1, max_order + 1):
    for i in range(0, len(segment) - order + 1):
      ngram = tuple(segment[i : i + order])
      ngram_counts[ngram] += 1
  return ngram_counts


def compute_bleu_matches(reference_corpus, translation_corpus, max_order=4):
  """Computes BLEU match stats of translations against one or more references.

  Args:
    reference_corpus: list of references for each translation. Each reference
      should be tokenized into a list of tokens.
    translation_corpus: list of translations to score. Each translation should
      be tokenized into a list of tokens.
    max_order: Maximum n-gram order to use when computing BLEU score.

  Returns:
    Aggregated n-gram stats for BLEU calculation.
  """
  reference_length = 0
  translation_length = 0
  bp = 1.0
  geo_mean = 0

  matches_by_order = [0] * max_order
  possible_matches_by_order = [0] * max_order
  precisions = []

  for references, translations in zip(reference_corpus, translation_corpus):
    reference_length += len(references)
    translation_length += len(translations)
    ref_ngram_counts = _get_ngrams(references, max_order)
    translation_ngram_counts = _get_ngrams(translations, max_order)

    overlap = {
        ngram: min(count, translation_ngram_counts[ngram])
        for ngram, count in ref_ngram_counts.items()
    }

    for ngram in overlap:
      matches_by_order[len(ngram) - 1] += overlap[ngram]
    for ngram in translation_ngram_counts:
      possible_matches_by_order[len(ngram) - 1] += translation_ngram_counts[
          ngram
      ]

  return (
      np.array(matches_by_order),
      np.array(possible_matches_by_order),
      np.array(reference_length),
      np.array(translation_length),
  )


def bleu_partial(ref_lines, hyp_lines, case_sensitive=False):
  """Compute n-gram statistics for two lists of references and translations."""
  if len(ref_lines) != len(hyp_lines):
    raise ValueError(
        "Reference and translation lists have different numbers of lines."
    )
  if not case_sensitive:
    ref_lines = [x.lower() for x in ref_lines]
    hyp_lines = [x.lower() for x in hyp_lines]
  ref_tokens = [bleu_tokenize(x) for x in ref_lines]
  hyp_tokens = [bleu_tokenize(x) for x in hyp_lines]
  return compute_bleu_matches(ref_tokens, hyp_tokens)


def complete_bleu(
    matches_by_order,
    possible_matches_by_order,
    reference_length,
    translation_length,
    max_order=4,
    use_bp=True,
):
  """Compute BLEU score from aggregated n-gram statistics."""
  precisions = [0] * max_order
  smooth = 1.0
  geo_mean = 0.0
  for i in range(0, max_order):
    if possible_matches_by_order[i] > 0:
      precisions[i] = matches_by_order[i] / possible_matches_by_order[i]
      if matches_by_order[i] > 0:
        precisions[i] = matches_by_order[i] / possible_matches_by_order[i]
      else:
        smooth *= 2
        precisions[i] = 1.0 / (smooth * possible_matches_by_order[i])
    else:
      precisions[i] = 0.0

  if max(precisions) > 0:
    p_log_sum = sum(math.log(p) for p in precisions if p)
    geo_mean = math.exp(p_log_sum / max_order)

  if use_bp:
    if not reference_length:
      bp = 1.0
    else:
      ratio = translation_length / reference_length
      if ratio <= 0.0:
        bp = 0.0
      elif ratio >= 1.0:
        bp = 1.0
      else:
        bp = math.exp(1 - 1.0 / ratio)
  bleu = geo_mean * bp
  return float(bleu) * 100.0


def bleu_local(ref_lines, hyp_lines, case_sensitive=False):
  """Compute BLEU for two lists of reference and hypothesis translations."""
  stats = bleu_partial(ref_lines, hyp_lines, case_sensitive=case_sensitive)
  return complete_bleu(*stats) * 100



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Transforms tests."""

from functools import partial
import operator
from typing import Any, Callable, Dict, Sequence
import unittest

from absl.testing import absltest, parameterized
from flax import errors
from flax import linen as nn
from flax import serialization
from flax import struct
from flax.core import copy, freeze, AxisMetadata
from flax.linen.transforms import _HashableProxy
import jax
from jax import random
import jax.numpy as jnp
import numpy as np

# TODO(jakevdp): use jax.debug_key_reuse directly once min jax version is 0.4.26
jax_debug_key_reuse = (jax.debug_key_reuse if hasattr(jax, 'debug_key_reuse')
                       else jax.enable_key_reuse_checks)

# Parse absl flags test_srcdir and test_tmpdir.
jax.config.parse_flags_with_absl()

# pylint: disable=attribute-defined-outside-init,unused-variable,g-wrong-blank-lines,g-bare-generic


def tree_equals(x, y):
  return jax.tree_util.tree_all(jax.tree_util.tree_map(operator.eq, x, y))


def tree_allclose(x, y):
  return jax.tree_util.tree_all(
      jax.tree_util.tree_map(lambda x, y: np.all(np.isclose(x, y)), x, y)
  )


id_fn = lambda x: x


class TransformedMLP(nn.Module):
  features: Sequence[int]
  transform: Callable = id_fn

  @nn.compact
  def __call__(self, inputs):
    x = inputs
    for i, feat in enumerate(self.features):
      # JIT the Module (it's __call__ fn by default.)
      x = self.transform(nn.Dense)(feat, name=f'layers_{i}')(x)
      if i != len(self.features) - 1:
        x = nn.relu(x)
    return x


def decorated_MLP(transform: Callable = id_fn):
  class MLP(nn.Module):
    features: Sequence[int]

    @transform
    @nn.compact
    def __call__(self, inputs):
      x = inputs
      for i, feat in enumerate(self.features):
        # JIT the Module (it's __call__ fn by default.)
        x = nn.Dense(feat, name=f'layers_{i}')(x)
        if i != len(self.features) - 1:
          x = nn.relu(x)
      return x

  return MLP


class TransformTest(parameterized.TestCase):

  def test_jit(self):
    key1, key2 = random.split(random.key(3), 2)
    x = random.uniform(key1, (4, 4))

    normal_model = TransformedMLP(features=[3, 4, 5])
    jit_model = TransformedMLP(features=[3, 4, 5], transform=nn.jit)
    init_variables = normal_model.init(key2, x)
    y1 = normal_model.apply(init_variables, x)
    y2 = jit_model.apply(init_variables, x)

    self.assertTrue(np.all(y1 == y2))

  def test_jit_decorated(self):
    key1, key2 = random.split(random.key(3), 2)
    x = random.uniform(key1, (4, 4))

    normal_model = decorated_MLP()(features=[3, 4, 5])
    jit_model = decorated_MLP(nn.jit)(features=[3, 4, 5])
    init_variables = normal_model.init(key2, x)
    y1 = normal_model.apply(init_variables, x)
    y2 = jit_model.apply(init_variables, x)

    self.assertTrue(np.all(y1 == y2))

  def test_remat(self):
    key1, key2 = random.split(random.key(3), 2)
    x = random.uniform(key1, (4, 4))

    normal_model = TransformedMLP(features=[3, 4, 5])
    remat_model = TransformedMLP(features=[3, 4, 5], transform=nn.remat)
    init_variables = normal_model.init(key2, x)
    y1 = normal_model.apply(init_variables, x)
    y2 = remat_model.apply(init_variables, x)

    self.assertTrue(np.all(y1 == y2))

  def test_remat_decorated(self):
    key1, key2 = random.split(random.key(3), 2)
    x = random.uniform(key1, (4, 4))

    normal_model = decorated_MLP()(features=[3, 4, 5])
    remat_model = decorated_MLP(nn.remat)(features=[3, 4, 5])
    init_variables = normal_model.init(key2, x)
    y1 = normal_model.apply(init_variables, x)
    y2 = remat_model.apply(init_variables, x)

    self.assertTrue(np.all(y1 == y2))

  def test_remat_kwargs(self):
    raise unittest.SkipTest('test breaks with grad')

    class ConditionalReLU(nn.Module):

      @nn.compact
      def __call__(self, input, apply_relu: bool = False):
        return nn.relu(input) if apply_relu else input

    key = random.key(0)
    x = jnp.ones((4, 4)) * -1
    remat_model = nn.remat(ConditionalReLU)()
    p = remat_model.init(key, x)
    y = remat_model.apply(p, x, apply_relu=True)

    self.assertTrue(np.all(y == jnp.zeros_like(x)))

    # This next line crashes with a concretization error
    _ = jax.grad(lambda x: remat_model.apply(p, x, apply_relu=True))(x)

  def test_remat_static_argnums(self):
    test = self

    class Foo(nn.Module):
      train_is_static: bool

      @nn.compact
      def __call__(self, inputs, train: bool):
        if self.train_is_static:
          test.assertTrue(isinstance(train, bool))
        else:
          test.assertTrue(isinstance(train, jnp.ndarray))

        return nn.Dense(3, use_bias=False)(inputs)

    # set train as a static argument
    FooRemat = nn.remat(Foo, static_argnums=(2,))
    foo = FooRemat(train_is_static=True)

    x = jnp.empty((1, 2))
    variables = foo.init(random.key(0), x, True)
    y = foo.apply(variables, x, False)
    self.assertEqual(y.shape, (1, 3))

    # set train as a non-static arguments
    FooRemat = nn.remat(Foo, static_argnums=())
    foo = FooRemat(train_is_static=False)

    variables = foo.init(random.key(0), x, True)
    y = foo.apply(variables, x, False)
    self.assertEqual(y.shape, (1, 3))

  def test_remat_decorator_static_argnums(self):
    test = self

    class FooTrainStatic(nn.Module):

      @partial(nn.remat, static_argnums=(2,))
      @nn.compact
      def __call__(self, inputs, train: bool):
        test.assertTrue(isinstance(train, bool))

        return nn.Dense(3, use_bias=False)(inputs)

    # set train as a static argument
    foo = FooTrainStatic()

    x = jnp.empty((1, 2))
    variables = foo.init(random.key(0), x, True)
    y = foo.apply(variables, x, False)
    self.assertEqual(y.shape, (1, 3))

    class FooTrainDynamic(nn.Module):

      @partial(nn.remat, static_argnums=())
      @nn.compact
      def __call__(self, inputs, train: bool):
        test.assertTrue(isinstance(train, jnp.ndarray))

        return nn.Dense(3, use_bias=False)(inputs)

    # set train as a non-static arguments
    foo = FooTrainDynamic()

    variables = foo.init(random.key(0), x, True)
    y = foo.apply(variables, x, False)
    self.assertEqual(y.shape, (1, 3))

  def test_vmap(self):
    key1, key2, key3 = random.split(random.key(3), 3)
    x = random.uniform(key1, (4, 4))
    x2 = random.uniform(key2, (5, 4, 4))

    def vmap(cls):
      return nn.vmap(
          cls,
          in_axes=(0,),
          variable_axes={'params': None},
          split_rngs={'params': False},
      )

    normal_model = TransformedMLP(features=[3, 4, 5])
    vmap_model = TransformedMLP(features=[3, 4, 5], transform=vmap)
    init_variables = normal_model.init(key3, x)
    # simulate vmap in python for comparison:
    y1 = jnp.vstack([
        normal_model.apply(init_variables, x2[i])[None, ...]
        for i in np.arange(x2.shape[0])
    ])
    y2 = vmap_model.apply(init_variables, x2)
    np.testing.assert_allclose(y1, y2, atol=1e-7)

  def test_vmap_decorated(self):
    key1, key2, key3 = random.split(random.key(3), 3)
    x = random.uniform(key1, (4, 4))
    x2 = random.uniform(key2, (5, 4, 4))

    def vmap(fn):
      return nn.vmap(
          fn,
          in_axes=(0,),
          variable_axes={'params': None},
          split_rngs={'params': False},
      )

    normal_model = decorated_MLP()(features=[3, 4, 5])
    vmap_model = decorated_MLP(vmap)(features=[3, 4, 5])
    init_variables = normal_model.init(key3, x)
    # simulate vmap in python for comparison:
    y1 = jnp.vstack([
        normal_model.apply(init_variables, x2[i])[None, ...]
        for i in np.arange(x2.shape[0])
    ])
    y2 = vmap_model.apply(init_variables, x2)
    np.testing.assert_allclose(y1, y2, atol=1e-7)

  def test_vmap_batchnorm(self):
    key1, key2, key3 = random.split(random.key(3), 3)
    x = random.uniform(key1, (4, 4))
    x2 = random.uniform(key2, (5, 4, 4))

    def vmap(cls):
      return nn.vmap(
          cls,
          in_axes=(0,),
          variable_axes={'params': None, 'batch_stats': None},
          split_rngs={'params': False},
          axis_name='batch',
      )

    class MlpBn(nn.Module):
      axis_name: Any = None

      @nn.compact
      def __call__(self, x):
        x = nn.Dense(3)(x)
        x = nn.BatchNorm(axis_name=self.axis_name, use_running_average=False)(x)
        return x

    normal_model = MlpBn()
    vmap_model = vmap(MlpBn)(axis_name='batch')
    init_variables = normal_model.init(key3, x)
    y1 = normal_model.apply(
        init_variables, x2.reshape((-1, 4)), mutable=['batch_stats']
    )[0]
    y1 = y1.reshape((5, 4, 3))
    y2 = vmap_model.apply(init_variables, x2, mutable=['batch_stats'])[0]
    np.testing.assert_allclose(y1, y2, atol=1e-5)

  def test_scan(self):
    class SimpleScan(nn.Module):
      features: int

      @nn.compact
      def __call__(self, c, xs):
        LSTM = nn.scan(
            nn.LSTMCell,
            variable_broadcast='params',
            split_rngs={'params': False},
        )
        return LSTM(self.features, name='lstm_cell')(c, xs)

    key1, key2 = random.split(random.key(0), 2)
    xs = random.uniform(key1, (5, 3, 2))
    dummy_rng = random.key(0)
    init_carry = nn.LSTMCell(2).initialize_carry(dummy_rng, xs[0].shape)
    model = SimpleScan(2)
    init_variables = model.init(key2, init_carry, xs)
    # simulate scan in python for comparison:
    c = init_carry
    ys = []
    lstmcell_variables = freeze(
        {'params': init_variables['params']['lstm_cell']}
    )
    for i in range(xs.shape[0]):
      c, y = nn.LSTMCell(2).apply(lstmcell_variables, c, xs[i])
      ys.append(y[None, ...])
    y1 = jnp.vstack(ys)

    c2, y2 = model.apply(init_variables, init_carry, xs)
    np.testing.assert_allclose(y1, y2, atol=1e-7)
    np.testing.assert_allclose(c[0], c2[0], atol=1e-7)
    np.testing.assert_allclose(c[1], c2[1], atol=1e-7)

  def test_scan_decorated(self):
    class SimpleScan(nn.Module):
      features: int

      @partial(
          nn.scan,
          variable_broadcast='params',
          in_axes=(nn.broadcast, 0),
          split_rngs={'params': False},
      )
      @nn.compact
      def __call__(self, c, b, xs):
        assert b.shape == (4,)
        return nn.LSTMCell(self.features, name='lstm_cell')(c, xs)

    key1, key2 = random.split(random.key(0), 2)
    xs = random.uniform(key1, (4, 3, 2))
    b = jnp.ones((4,))
    dummy_rng = random.key(0)
    init_carry = nn.LSTMCell(2).initialize_carry(dummy_rng, xs[0].shape)
    model = SimpleScan(2)
    init_variables = model.init(key2, init_carry, b, xs)
    # simulate scan in python for comparison:
    c = init_carry
    ys = []
    lstmcell_variables = freeze(
        {'params': init_variables['params']['lstm_cell']}
    )
    for i in range(xs.shape[0]):
      c, y = nn.LSTMCell(2).apply(lstmcell_variables, c, xs[i])
      ys.append(y[None, ...])
    y1 = jnp.vstack(ys)

    c2, y2 = model.apply(init_variables, init_carry, b, xs)
    np.testing.assert_allclose(y1, y2, atol=1e-7)
    np.testing.assert_allclose(c[0], c2[0], atol=1e-7)
    np.testing.assert_allclose(c[1], c2[1], atol=1e-7)

  def test_scan_negative_axes(self):
    class Foo(nn.Module):

      @nn.compact
      def __call__(self, _, x):
        x = nn.Dense(4)(x)
        return None, x

    class Bar(nn.Module):

      @nn.compact
      def __call__(self, x):
        _, x = nn.scan(
            Foo,
            variable_broadcast='params',
            split_rngs=dict(params=False),
            in_axes=1,
            out_axes=-1,
        )()(None, x)
        return x

    y, variables = Bar().init_with_output(
        {'params': jax.random.PRNGKey(0)},
        jax.random.normal(jax.random.PRNGKey(1), shape=[1, 2, 3]),
    )
    params = variables['params']

    self.assertEqual(y.shape, (1, 4, 2))
    self.assertEqual(params['ScanFoo_0']['Dense_0']['kernel'].shape, (3, 4))
    self.assertEqual(params['ScanFoo_0']['Dense_0']['bias'].shape, (4,))

  def test_multiscope_lifting_simple(self):
    class Counter(nn.Module):

      @nn.compact
      def __call__(self):
        v = self.variable('counter', 'foo', lambda: jnp.array([0]))
        v.value += jnp.array([1])
        return v.value

    class Outer(nn.Module):

      @nn.compact
      def __call__(self, x):
        cntr = nn.jit(Counter)(name='cntr')()
        return x

    class Inner(nn.Module):
      outer_module: nn.Module

      @nn.compact
      def __call__(self, x):
        return self.outer_module(x)

    class Test(nn.Module):

      @nn.compact
      def __call__(self, x):
        outer_dense = nn.jit(Outer)(name='outer')
        # we share stateful outer module as arg to two different, transformed modules:
        inner = nn.jit(Inner)(outer_dense, name='inner1')
        inner2 = nn.jit(Inner)(outer_dense, name='inner2')
        res = inner(x) + inner2(x)
        return res

    x = jnp.ones((10, 10))
    rngs = random.key(0)
    init_vars = Test(parent=None).init(rngs, x)
    _, new_vars = Test(parent=None).apply(init_vars, x, mutable=['counter'])
    self.assertEqual(
        init_vars['counter']['outer']['cntr']['foo'], jnp.array([2], jnp.int32)
    )
    self.assertEqual(
        new_vars['counter']['outer']['cntr']['foo'], jnp.array([4], jnp.int32)
    )

  def test_multiscope_lifting_simple_decorator(self):
    class Counter(nn.Module):

      @nn.jit
      @nn.compact
      def __call__(self):
        v = self.variable('counter', 'foo', lambda: jnp.array([0]))
        v.value += jnp.array([1])
        return v.value

    class Outer(nn.Module):

      @nn.jit
      @nn.compact
      def __call__(self, x):
        cntr = Counter(name='cntr')()
        return x

    class Inner(nn.Module):
      outer_module: nn.Module

      @nn.jit
      @nn.compact
      def __call__(self, x):
        return self.outer_module(x)

    class Test(nn.Module):

      @nn.compact
      def __call__(self, x):
        outer_dense = Outer(name='outer')
        # we share stateful outer module as arg to two different, transformed modules:
        inner = Inner(outer_dense, name='inner1')
        inner2 = Inner(outer_dense, name='inner2')
        res = inner(x) + inner2(x)
        return res

    x = jnp.ones((1, 1))
    rngs = random.key(0)
    init_vars = Test(parent=None).init(rngs, x)
    _, new_vars = Test(parent=None).apply(init_vars, x, mutable=['counter'])
    self.assertEqual(
        init_vars['counter']['outer']['cntr']['foo'], jnp.array([2], jnp.int32)
    )
    self.assertEqual(
        new_vars['counter']['outer']['cntr']['foo'], jnp.array([4], jnp.int32)
    )

  def test_multiscope_lifting_argtree(self):
    class Counter(nn.Module):

      @nn.compact
      def __call__(self):
        v = self.variable('counter', 'foo', lambda: jnp.array([0]))
        v.value += jnp.array([1])
        return v.value

    class Outer(nn.Module):

      @nn.compact
      def __call__(self, x):
        cntr = nn.jit(Counter)(name='cntr')()
        return x

    class Inner(nn.Module):
      outer_module: Sequence[nn.Module]

      @nn.compact
      def __call__(self, x):
        return self.outer_module[0](x) + self.outer_module[1](x)

    class Test(nn.Module):

      @nn.compact
      def __call__(self, x):
        outer_dense1 = nn.jit(Outer)(name='outer1')
        outer_dense2 = nn.jit(Outer)(name='outer2')
        # we share stateful outer module as arg to two different, transformed modules:
        inner1 = nn.jit(Inner)((outer_dense1, outer_dense2), name='inner1')
        inner2 = nn.jit(Inner)((outer_dense1, outer_dense2), name='inner2')
        res = inner1(x) + inner2(x)
        return res

    x = jnp.ones((1, 1))
    rngs = random.key(0)
    init_vars = Test(parent=None).init(rngs, x)
    _, new_vars = Test(parent=None).apply(init_vars, x, mutable=['counter'])
    self.assertEqual(
        init_vars['counter']['outer1']['cntr']['foo'], jnp.array([2], jnp.int32)
    )
    self.assertEqual(
        new_vars['counter']['outer1']['cntr']['foo'], jnp.array([4], jnp.int32)
    )
    self.assertEqual(
        init_vars['counter']['outer2']['cntr']['foo'], jnp.array([2], jnp.int32)
    )
    self.assertEqual(
        new_vars['counter']['outer2']['cntr']['foo'], jnp.array([4], jnp.int32)
    )

  def test_multiscope_lifting_argtree_decorator(self):
    class Counter(nn.Module):

      @nn.jit
      @nn.compact
      def __call__(self):
        v = self.variable('counter', 'foo', lambda: jnp.array([0]))
        v.value += jnp.array([1])
        return v.value

    class Outer(nn.Module):

      @nn.jit
      @nn.compact
      def __call__(self, x):
        cntr = nn.jit(Counter)(name='cntr')()
        return x

    class Inner(nn.Module):
      outer_module: Sequence[nn.Module]

      @nn.jit
      @nn.compact
      def __call__(self, x):
        return self.outer_module[0](x) + self.outer_module[1](x)

    class Test(nn.Module):

      @nn.compact
      def __call__(self, x):
        outer_dense1 = Outer(name='outer1')
        outer_dense2 = Outer(name='outer2')
        # we share stateful outer module as arg to two different, transformed modules:
        inner1 = Inner((outer_dense1, outer_dense2), name='inner1')
        inner2 = Inner((outer_dense1, outer_dense2), name='inner2')
        res = inner1(x) + inner2(x)
        return res

    x = jnp.ones((1, 1))
    rngs = random.key(0)
    init_vars = Test(parent=None).init(rngs, x)
    _, new_vars = Test(parent=None).apply(init_vars, x, mutable=['counter'])
    self.assertEqual(
        init_vars['counter']['outer1']['cntr']['foo'], jnp.array([2], jnp.int32)
    )
    self.assertEqual(
        new_vars['counter']['outer1']['cntr']['foo'], jnp.array([4], jnp.int32)
    )
    self.assertEqual(
        init_vars['counter']['outer2']['cntr']['foo'], jnp.array([2], jnp.int32)
    )
    self.assertEqual(
        new_vars['counter']['outer2']['cntr']['foo'], jnp.array([4], jnp.int32)
    )

  def test_multiscope_lifting_simple_decorator_w_jit(self):
    # TODO: actually test jaxpr on a simpler module.
    class Counter(nn.Module):

      @nn.jit
      @nn.compact
      def __call__(self):
        v = self.variable('counter', 'foo', lambda: jnp.array([0]))
        v.value += jnp.array([1])
        return v.value

    class Outer(nn.Module):

      @nn.jit
      @nn.compact
      def __call__(self, x):
        cntr = Counter(name='cntr')()
        return x

    class Inner(nn.Module):
      outer_module: nn.Module

      @nn.jit
      @nn.compact
      def __call__(self, x):
        return self.outer_module(x)

    class Test(nn.Module):

      @nn.jit
      @nn.compact
      def __call__(self, x):
        outer_dense = Outer(name='outer')
        # we share stateful outer module as arg to two different, transformed modules:
        inner = Inner(outer_dense, name='inner1')
        inner2 = Inner(outer_dense, name='inner2')
        res = inner(x) + inner2(x)
        return res

    x = jnp.ones((1, 1))
    rngs = random.key(0)
    init_vars = Test(parent=None).init(rngs, x)
    _, new_vars = Test(parent=None).apply(init_vars, x, mutable=['counter'])
    self.assertEqual(
        init_vars['counter']['outer']['cntr']['foo'], jnp.array([2], jnp.int32)
    )
    self.assertEqual(
        new_vars['counter']['outer']['cntr']['foo'], jnp.array([4], jnp.int32)
    )

  def test_vmapped_outer_module(self):
    class Outer(nn.Module):

      @nn.jit
      @nn.compact
      def __call__(self, x):
        return nn.Dense(5)(x)

    class Inner(nn.Module):
      outer_module: nn.Module

      @partial(
          nn.vmap,
          in_axes=(0,),
          variable_axes={'params': 0},
          split_rngs={'params': True},
      )
      @nn.jit
      @nn.compact
      def __call__(self, x):
        return self.outer_module(x)

    class Test(nn.Module):

      @nn.compact
      def __call__(self, x):
        outer_dense = Outer(name='outer')
        inner = Inner(outer_dense, name='inner1')
        inner2 = Inner(outer_dense, name='inner2')
        res = inner(x) + inner2(x)
        return res

    x = jnp.ones((3, 1, 2))
    rngs = random.key(0)
    init_vars = Test(parent=None).init(rngs, x)
    y = Test(parent=None).apply(init_vars, x)
    self.assertEqual(
        init_vars['params']['outer']['Dense_0']['kernel'].shape, (3, 2, 5)
    )
    self.assertEqual(
        init_vars['params']['outer']['Dense_0']['bias'].shape, (3, 5)
    )
    self.assertEqual(y.shape, (3, 1, 5))

  def test_module_transform_with_setup(self):
    class Foo(nn.Module):

      def setup(self):
        self.test = self.param('test', nn.initializers.ones_init(), ())

      def __call__(self, x):
        return x * self.test

    FooVmap = nn.vmap(
        Foo,
        in_axes=0,
        out_axes=0,
        variable_axes={'params': 0},
        split_rngs={'params': True},
    )
    variables = FooVmap().init(random.key(0), jnp.ones((4,)))
    self.assertEqual(variables['params']['test'].shape, (4,))

  def test_nested_module_args_vmap(self):
    class A(nn.Module):

      @nn.compact
      def __call__(self, x):
        return nn.Dense(3)(x)

    class B(nn.Module):
      A: nn.Module

      @nn.compact
      def __call__(self, x):
        return self.A(x)

    class C(nn.Module):
      B: nn.Module

      @partial(
          nn.vmap, variable_axes={'params': 0}, split_rngs={'params': True}
      )
      @nn.compact
      def __call__(self, x):
        return self.B(x)

    class D(nn.Module):

      @nn.compact
      def __call__(self, x):
        a = A()
        b = B(a)
        c = C(b)
        return c(x)

    key = random.key(0)
    x = jnp.ones((10, 10))
    p = D().init(key, x)

    variable_shapes = jax.tree_util.tree_map(jnp.shape, p)
    self.assertEqual(
        variable_shapes['params']['A_0']['Dense_0']['kernel'], (10, 10, 3)
    )
    self.assertEqual(
        variable_shapes['params']['A_0']['Dense_0']['bias'], (10, 3)
    )

  def test_nested_module_args_vmap_2(self):
    class A(nn.Module):

      @nn.compact
      def __call__(self, x):
        return nn.Dense(3)(x)

    class B(nn.Module):
      A: nn.Module

      @nn.compact
      def __call__(self, x):
        return self.A(x)

    class C(nn.Module):
      A: nn.Module
      B: nn.Module

      @partial(
          nn.vmap, variable_axes={'params': 0}, split_rngs={'params': True}
      )
      @nn.compact
      def __call__(self, x):
        return self.B(x) + self.A(x)

    class D(nn.Module):

      @nn.compact
      def __call__(self, x):
        a1 = A()
        a2 = A()
        b = B(a1)
        c = C(a2, b)
        return c(x)

    key = random.key(0)
    x = jnp.ones((10, 10))
    p = D().init(key, x)

    variable_shapes = jax.tree_util.tree_map(jnp.shape, p)
    self.assertEqual(
        variable_shapes['params']['A_0']['Dense_0']['kernel'], (10, 10, 3)
    )
    self.assertEqual(
        variable_shapes['params']['A_0']['Dense_0']['bias'], (10, 3)
    )
    self.assertEqual(
        variable_shapes['params']['A_1']['Dense_0']['kernel'], (10, 10, 3)
    )
    self.assertEqual(
        variable_shapes['params']['A_1']['Dense_0']['bias'], (10, 3)
    )

  def test_nested_setup_calls_count(self):
    D = 3
    N = 4
    setup_cntr = 0
    call_cntr = 0

    class Repeat(nn.Module):
      mdl_def: Any

      def setup(self):
        self.lyrs = [self.mdl_def() for _ in range(N)]

      @nn.remat  # we just use remat as a convenient test of transform logic
      def __call__(self, x):
        for lyr in self.lyrs:
          lyr(x)
        return x

    class Counter(nn.Module):

      def setup(self):
        nonlocal setup_cntr
        setup_cntr += 1
        self.dense = nn.Dense(2, use_bias=False)

      @nn.remat
      def __call__(self, x):
        nonlocal call_cntr
        call_cntr += 1
        return self.dense(x)

    def nested_repeat(mdl):
      for _ in range(D):
        mdl = partial(Repeat, mdl)
      return mdl()

    _ = nested_repeat(Counter).init(random.key(0), jnp.ones((2,)))
    # setup_cntr == 128 due to 1 call in Counter.setup by _validate_setup
    # and 1 further "real" call.
    self.assertEqual(setup_cntr, 128)
    self.assertEqual(call_cntr, 64)

  def test_multimethod_setup_calls(self):
    cntr = 0

    class A(nn.Module):

      def setup(self):
        nonlocal cntr
        cntr += 1
        self.d = nn.Dense(2)

      @nn.remat
      def foo(self, x):
        return self.d(x)

      @nn.remat
      def bar(self, x):
        return self.d(x)

    class B(nn.Module):

      def setup(self):
        self.a = A()

      def __call__(self, x):
        y1 = self.a.foo(x)
        y2 = self.a.bar(x)
        return y1, y2

    key = random.key(0)
    x = jnp.ones((2,))
    (y1, y2), _ = B().init_with_output(key, x)
    np.testing.assert_array_equal(y1, y2)
    # cntr == 4 due to:
    # 1 call by _validate_setup
    # 1 call for the setup() outside transform boundary
    # and two further "real" calls in transform boundaries
    self.assertEqual(cntr, 4)

  def test_toplevel_submodule_adoption_transform(self):
    class A(nn.Module):

      @nn.compact
      def __call__(self, x):
        return nn.Dense(3)(x)

    class B(nn.Module):
      A: nn.Module

      @nn.compact
      def __call__(self, x):
        return self.A(x)

    class C(nn.Module):
      A: nn.Module
      B: nn.Module

      @partial(
          nn.vmap, variable_axes={'params': 0}, split_rngs={'params': True}
      )
      @nn.compact
      def __call__(self, x):
        return self.B(x) + self.A(x)

    class Csimple(nn.Module):
      A: nn.Module
      B: nn.Module

      @nn.compact
      def __call__(self, x):
        return self.B(x) + self.A(x)

    class D(nn.Module):

      @nn.compact
      def __call__(self, x):
        a1 = A()
        a2 = A()
        b = B(a1)
        c = C(a2, b)
        return c(x)

    key = random.key(0)
    x = jnp.ones((10, 10))
    p1 = D().init(key, x)
    y1 = D().apply(p1, x)

    a1 = A()
    a2 = A()
    b = B(a1)
    p2 = freeze({
        'params': {
            'A': p1['params']['A_0'],
            'B': {
                'A': p1['params']['A_1'],
            },
        }
    })

    # Test method wrapper transform.
    y2 = C(a2, b).apply(p2, x)
    np.testing.assert_allclose(y1, y2, atol=1e-7)
    # Test class transform.
    Ctrafo = nn.vmap(
        Csimple, variable_axes={'params': 0}, split_rngs={'params': True}
    )

    y3 = Ctrafo(a2, b).apply(p2, x)
    np.testing.assert_allclose(y1, y3, atol=1e-7)

  def test_toplevel_submodule_adoption_pytree_transform(self):
    class A(nn.Module):

      @nn.compact
      def __call__(self, c, x):
        counter = self.variable('counter', 'i', jnp.zeros, ())
        counter.value += 1
        x = nn.Dense(1)(x)
        return c, x

    class B(nn.Module):
      A: Any

      @nn.compact
      def __call__(self, c, x):
        return self.A['foo'](*self.A['bar'](c, x))

    a = A()
    As = {'foo': A(), 'bar': A()}
    b = nn.scan(
        B,
        in_axes=0,
        variable_carry='counter',
        variable_broadcast='params',
        split_rngs={'params': False},
    )(As)

    key = random.key(0)
    x = jnp.ones((10, 2))

    p = B(As).init(key, x, x)
    y, cntrs = b.apply(p, x, x, mutable='counter')
    ref_cntrs = {
        'counter': {
            'A_bar': {
                'i': jnp.array(11.0),
            },
            'A_foo': {
                'i': jnp.array(11.0),
            },
        },
    }
    self.assertTrue(
        jax.tree_util.tree_all(
            jax.tree_util.tree_map(
                lambda x, y: np.testing.assert_allclose(x, y, atol=1e-7),
                cntrs,
                ref_cntrs,
            )
        )
    )

  def test_partially_applied_module_constructor_transform(self):
    k = random.key(0)
    x = jnp.ones((3, 4, 4))
    dense = partial(nn.Dense, use_bias=False)
    vmap_dense = nn.vmap(
        dense, variable_axes={'params': 0}, split_rngs={'params': True}
    )(4)
    init_vars = vmap_dense.init(k, x)
    init_vars_shapes = jax.tree_util.tree_map(jnp.shape, init_vars)
    ref_var_shapes = {
        'params': {
            'kernel': (3, 4, 4),
        },
    }
    self.assertTrue(tree_equals(init_vars_shapes, ref_var_shapes))

  def test_partial_module_method(self):
    k = random.key(0)
    x = jnp.ones((3, 4, 4))

    class Foo(nn.Module):

      @nn.compact
      def inner(self, x):
        return nn.Dense(2, use_bias=False)(x)

      def __call__(self, x):
        return nn.vmap(
            partial(Foo.inner),
            variable_axes={'params': 0},
            split_rngs={'params': True},
        )(self, x)

    init_vars = Foo().init(k, x)
    init_vars_shapes = jax.tree_util.tree_map(jnp.shape, init_vars)
    ref_var_shapes = {
        'params': {'Dense_0': {'kernel': (3, 4, 2)}},
    }
    self.assertTrue(tree_equals(init_vars_shapes, ref_var_shapes))

  def test_variable_in_args_transform(self):
    class Test(nn.Module):

      @nn.jit
      @nn.compact
      def __call__(self, x):
        baz = self.variable('test', 'baz', jnp.zeros, x.shape)
        y = self.mutate_variable_in_method(x, baz)
        return y

      @nn.jit
      def mutate_variable_in_method(self, x, baz):
        baz.value += x
        return baz.value

    k = random.key(0)
    x = jnp.ones((1,))
    variables = Test().init(k, x)
    np.testing.assert_allclose(
        variables['test']['baz'],
        jnp.array([
            1.0,
        ]),
        atol=1e-7,
    )
    y, variables = Test().apply(variables, x, mutable=['test'])
    np.testing.assert_allclose(
        variables['test']['baz'],
        jnp.array([
            2.0,
        ]),
        atol=1e-7,
    )

  def test_module_instance_in_args_transform(self):
    class Inner(nn.Module):

      @nn.jit
      @nn.compact
      def __call__(self, x):
        baz = self.variable('test', 'baz', jnp.zeros, x.shape)
        baz.value += x
        return baz.value

    class Test(nn.Module):

      @nn.jit
      @nn.compact
      def __call__(self, x):
        inner = Inner(name='inner')
        y = self.call_instance_arg_in_method(x, inner)
        return y

      @nn.jit
      def call_instance_arg_in_method(self, x, inner):
        return inner(x)

    k = random.key(0)
    x = jnp.ones((1,))
    variables = Test().init(k, x)
    np.testing.assert_allclose(
        variables['test']['inner']['baz'],
        jnp.array([
            1.0,
        ]),
        atol=1e-7,
    )
    y, variables = Test().apply(variables, x, mutable=['test'])
    np.testing.assert_allclose(
        variables['test']['inner']['baz'],
        jnp.array([
            2.0,
        ]),
        atol=1e-7,
    )

  def test_module_instance_in_args_transform_nested(self):
    class Inner(nn.Module):

      @nn.jit
      @nn.compact
      def __call__(self, x):
        baz = self.variable('test', 'baz', jnp.zeros, x.shape)
        baz.value += x
        return baz.value

    class Outer(nn.Module):

      @nn.jit
      @nn.compact
      def __call__(self, inner, x):
        y = self.call_instance_arg_in_method(x, inner)
        return y

      @nn.jit
      def call_instance_arg_in_method(self, x, inner):
        return inner(x)

    class Test(nn.Module):

      @nn.jit
      @nn.compact
      def __call__(self, x):
        inner = Inner(name='inner')
        outer = Outer(name='outer')
        return outer(inner, x)

    k = random.key(0)
    x = jnp.ones((1,))
    variables = Test().init(k, x)
    np.testing.assert_allclose(
        variables['test']['inner']['baz'],
        jnp.array([
            1.0,
        ]),
        atol=1e-7,
    )
    y, variables = Test().apply(variables, x, mutable=['test'])
    np.testing.assert_allclose(
        variables['test']['inner']['baz'],
        jnp.array([
            2.0,
        ]),
        atol=1e-7,
    )

  def test_nested_variable_passing(self):
    class NestedVarUser(nn.Module):
      somevar: nn.Variable

      @nn.jit
      @nn.compact
      def __call__(self, x):
        self.somevar.value += x
        return x

    class VarUser(nn.Module):
      somevar: nn.Variable

      @nn.jit
      @nn.compact
      def __call__(self, x):
        return NestedVarUser(self.somevar)(x)

    class VarPasser(nn.Module):

      @nn.jit
      @nn.compact
      def __call__(self, x):
        baz = self.variable('test', 'baz', jnp.zeros, x.shape)
        y = VarUser(baz)(x)
        return y

    k = random.key(0)
    x = jnp.ones((1,))
    variables = VarPasser().init(k, x)
    np.testing.assert_allclose(
        variables['test']['baz'],
        jnp.array([
            1.0,
        ]),
        atol=1e-7,
    )
    y, variables = VarPasser().apply(variables, x, mutable=['test'])
    np.testing.assert_allclose(
        variables['test']['baz'],
        jnp.array([
            2.0,
        ]),
        atol=1e-7,
    )

  def test_returned_module_warning(self):
    class Foo(nn.Module):

      @nn.compact
      def __call__(self, x):
        return x

    class Bar(nn.Module):

      @nn.compact
      def __call__(self, x):
        f = self._helper()
        return f(x)

      @nn.jit
      def _helper(self):
        return Foo()

    b = Bar()
    with self.assertRaises(errors.TransformedMethodReturnValueError):
      b.apply({}, jnp.ones(2))

  def test_returned_variable_warning(self):
    class Bar(nn.Module):

      @nn.compact
      def __call__(self, x):
        f = self._helper()
        return f(x)

      @nn.jit
      def _helper(self):
        return nn.Variable(None, None, None, False)

    b = Bar()
    with self.assertRaises(errors.TransformedMethodReturnValueError):
      b.apply({}, jnp.ones(2))

  def test_nowrap(self):
    class Bar(nn.Module):

      @nn.compact
      def __call__(self, x):
        return self._helper(x)

      @nn.nowrap
      def _helper(self, x):
        if len(nn.module._context.module_stack) > 2:  # pylint: disable=protected-access
          raise ValueError('Module stack too deep.')
        return x

    b = Bar()
    b.apply({}, jnp.ones(2))

  def test_map_variables_tied_autoencoder(self):
    def trans(variables):
      return jax.tree_util.tree_map(lambda x: x.T, variables)

    class TiedAutencoder(nn.Module):
      features: int
      latents: int

      @nn.compact
      def _call(self, x, decode):
        def f(self):
          return nn.Dense(
              self.features if decode else self.latents, use_bias=False
          )(x)

        if decode:
          map_fn = trans
        else:
          map_fn = lambda x: x
        return nn.map_variables(f, 'params', map_fn, map_fn, mutable=True)(self)

      def encode(self, x):
        return self._call(x, False)

      def decode(self, x):
        return self._call(x, True)

      def __call__(self, x):
        return self.decode(self.encode(x))

    x = jnp.ones((2, 4))
    ae = TiedAutencoder(4, 5)
    variables = ae.init(random.key(0), x)
    param_shapes = jax.tree_util.tree_map(jnp.shape, variables['params'])
    self.assertEqual(param_shapes, {'Dense_0': {'kernel': (4, 5)}})

  def test_map_variables_bit_weights(self):
    class BitWeights(nn.Module):

      @nn.compact
      def __call__(self, x):
        def sign(x):
          return jax.tree_util.tree_map(jnp.sign, x)

        BitDense = nn.map_variables(nn.Dense, 'params', sign, init=True)
        return BitDense(4)(x)

    bw = BitWeights()
    x = jnp.ones((2, 4))
    y, variables = bw.init_with_output(random.key(0), x)
    y_2 = bw.apply(variables, x)
    np.testing.assert_allclose(y, y_2)

  def test_remat_scan(self):
    class BigModel(nn.Module):

      @nn.compact
      def __call__(self, x):
        DenseStack = nn.remat_scan(nn.Dense, lengths=(100,))
        return DenseStack(8, name='dense_stack')(x)

    x = jnp.ones((2, 8))
    model = BigModel()
    variables = model.init(random.key(0), x)
    param_shapes = jax.tree_util.tree_map(jnp.shape, variables['params'])
    self.assertEqual(param_shapes['dense_stack']['kernel'], (100, 8, 8))
    self.assertEqual(param_shapes['dense_stack']['bias'], (100, 8))
    y = model.apply(variables, x)
    self.assertEqual(y.shape, (2, 8))

  def test_vjp(self):
    class Bar(nn.Module):

      @nn.compact
      def __call__(self, x, y):
        p = self.param('test', nn.initializers.constant(0.5), ())
        self.variable('state', 'counter', lambda: 0)
        return p * x * y

    class Foo(nn.Module):

      @nn.compact
      def __call__(self, x, y):
        z, bwd = nn.vjp(Bar.__call__, Bar(), x, y)
        return bwd(jnp.ones(z.shape))

    x = jnp.array([1.0, 2.0, 3.0])
    y = jnp.array([4.0, 5.0, 6.0])
    params = Foo().init(random.key(0), x, y)
    params_grad, x_grad, y_grad = Foo().apply(params, x, y)
    self.assertEqual(
        params_grad,
        {
            'params': nn.FrozenDict({'test': 32.0}),
        },
    )
    np.testing.assert_allclose(x_grad, [2.0, 2.5, 3.0])
    np.testing.assert_allclose(y_grad, [0.5, 1.0, 1.5])

  def test_jvp(self):
    class Bar(nn.Module):

      @nn.compact
      def __call__(self, x):
        p = self.param('test', nn.initializers.zeros, ())
        self.variable('state', 'counter', lambda: 0)
        return p * x

    class Foo(nn.Module):

      @nn.compact
      def __call__(self, x):
        bar = Bar()
        vars_t = jax.tree_util.tree_map(
            jnp.ones_like, bar.variables.get('params', {})
        )
        _, out_t = nn.jvp(
            Bar.__call__, bar, (x,), (jnp.zeros_like(x),), {'params': vars_t}
        )
        return out_t

    x = jnp.ones((3,))
    params = Foo().init(random.key(0), x)
    y_t = Foo().apply(params, x)
    np.testing.assert_allclose(y_t, jnp.ones_like(x))

  def test_complicated_alias_mutation(self):
    class A(nn.Module):
      b: nn.Module

      @nn.jit
      @nn.compact
      def __call__(self, x):
        return self.b(x)

    class B(nn.Module):
      c: nn.Module

      @nn.jit
      @nn.compact
      def __call__(self, x):
        y = C(name='outer_c')(x)
        z = self.c(x)
        return z

    class C(nn.Module):

      @nn.jit
      @nn.compact
      def __call__(self, x):
        initialized = self.has_variable('muts', 'v')
        v = self.variable('muts', 'v', lambda: jnp.zeros_like(x))
        if initialized:
          v.value += x
        return x

    a = A(b=B(c=C()))
    k = random.key(0)
    x = jnp.ones((1,), jnp.float32)
    vs = a.init(k, x)
    y, vs_new = a.apply(
        vs,
        x,
        mutable=[
            'muts',
        ],
    )
    np.testing.assert_array_equal(
        vs_new['muts']['b']['c']['v'], jnp.array([1.0], jnp.float32)
    )
    np.testing.assert_array_equal(
        vs_new['muts']['b']['outer_c']['v'], jnp.array([1.0], jnp.float32)
    )

  def test_custom_vjp(self):
    class Foo(nn.Module):

      @nn.compact
      def __call__(self, x):
        def f(mdl, x):
          return mdl(x)

        def fwd(mdl, x):
          return nn.vjp(f, mdl, x)

        def bwd(vjp_fn, y_t):
          params_t, input_t = vjp_fn(y_t)
          params_t = jax.tree_util.tree_map(jnp.sign, params_t)
          return params_t, input_t

        sign_grad = nn.custom_vjp(f, forward_fn=fwd, backward_fn=bwd)
        return sign_grad(nn.Dense(1), x).reshape(())

    x = jnp.ones((2,))
    variables = Foo().init(random.key(0), x)
    grad = jax.grad(Foo().apply)(variables, x)
    for grad_leaf in jax.tree_util.tree_leaves(grad):
      self.assertTrue(jnp.all(jnp.abs(grad_leaf) == 1.0))

  def test_transform_with_setup_and_methods_on_submodules(self):
    # This is the archetypal example motivating the introduction of
    # SetupState as a triple-enum to handle multiple setup() calls
    # across transform boundaries and scope reuse.
    class Foo(nn.Module):

      def setup(self):
        self.inner = nn.Dense(2)

      def helper(self, x, m):
        return m(x)

      def __call__(self, x):
        return self.helper(x, self.inner)

    k = random.key(0)
    x = jnp.ones((2,))
    vs_foo = Foo().init(k, x)

    class Bar(nn.Module):

      def setup(self):
        self.inner = nn.Dense(2)

      @nn.jit
      def helper(self, x, m):
        return m(x)

      @nn.jit
      def __call__(self, x):
        return self.helper(x, self.inner)

    vs_bar = Bar().init(k, x)
    self.assertTrue(
        tree_equals(
            jax.tree_util.tree_map(jnp.shape, vs_foo),
            jax.tree_util.tree_map(jnp.shape, vs_bar),
        )
    )

  def test_transform_methods_on_submodules_still_reserve_names(self):
    class Foo(nn.Module):

      @nn.jit
      def helper(self, x, m):
        conflicting_a = nn.Dense(2, name='a')
        return m(x)

      @nn.jit
      @nn.compact
      def __call__(self, x):
        a = nn.Dense(2, name='a')
        return self.helper(x, a)

    k = random.key(0)
    x = jnp.ones((2,))
    with self.assertRaises(errors.NameInUseError):
      vs = Foo().init(k, x)

  def test_transform_setup_still_reserve_names(self):
    class Identity(nn.Module):

      @nn.compact
      def __call__(self, x):
        return x

    class Test(nn.Module):

      def setup(self):
        self.sub = Identity()
        self.sub = Identity()

      @nn.jit
      def __call__(self, x):
        return x

    k = random.key(0)
    x = jnp.array([1.0])

    with self.assertRaises(errors.NameInUseError):
      y = Test().init(k, x)

  def test_transform_with_setup_and_methods_on_submodule_pytrees(self):
    class Foo(nn.Module):

      def setup(self):
        self.inners = [nn.Dense(2), nn.Dense(2)]

      def helper(self, x, ms):
        return ms[0](x) + ms[1](x)

      def __call__(self, x):
        return self.helper(x, self.inners)

    class JitFoo(nn.Module):

      def setup(self):
        self.inners = [nn.Dense(2), nn.Dense(2)]

      @nn.jit
      def helper(self, x, ms):
        return ms[0](x) + ms[1](x)

      @nn.jit
      def __call__(self, x):
        return self.helper(x, self.inners)

    k = random.key(0)
    x = jnp.ones((2,))

    vs_0 = Foo().init(k, x)
    vs_1 = JitFoo().init(k, x)

    self.assertTrue(tree_allclose(vs_0, vs_1))

  def test_transform_setup_still_reserve_names_pytrees(self):
    class Identity(nn.Module):

      @nn.compact
      def __call__(self, x):
        return x

    class Test(nn.Module):

      def setup(self):
        self.subs = [Identity(), Identity()]
        self.subs = [Identity(), Identity()]

      @nn.jit
      def __call__(self, x):
        return x

    k = random.key(0)
    x = jnp.array([1.0])

    msg = r'Could not create submodule "subs_0".*'
    with self.assertRaisesRegex(errors.NameInUseError, msg):
      y = Test().init(k, x)

  def test_scan_of_setup_parameter(self):
    class Body(nn.Module):

      def setup(self):
        self.dense = nn.Dense(1)
        self.p = self.param('p', lambda k: jnp.ones((1,)))

      def __call__(self, x):
        return self.dense(x) + self.p, None

    scanbody = nn.scan(
        Body, variable_axes={'params': 0}, split_rngs={'params': True}, length=2
    )
    k = random.key(0)
    x = jnp.ones((1,))
    vs = scanbody().init(k, x)
    y = scanbody().apply(vs, x)

  def test_multi_method_class_transform(self):
    class Foo(nn.Module):

      def setup(self):
        self.dense0 = nn.Dense(2)
        self.dense1 = nn.Dense(2)

      def method_0(self, x):
        return self.dense0(x), x

      def method_1(self, x, y):
        return self.dense1(x) + y, None

    class Bar(nn.Module):

      @nn.compact
      def __call__(self, x):
        ScanFoo = nn.scan(
            Foo,
            methods={
                'method_0': dict(
                    variable_axes={'params': 0},
                    split_rngs={'params': True},
                    in_axes=nn.broadcast,
                    out_axes=0,
                    length=3,
                ),
                'method_1': dict(
                    variable_axes={'params': 0},
                    split_rngs={'params': True},
                    in_axes=0,
                    length=3,
                ),
            },
        )
        sf = ScanFoo()
        y, ys = sf.method_0(x)
        z, _ = sf.method_1(y, ys)
        return z

    k = random.key(0)
    x = random.uniform(random.key(1), (2, 2))
    vs = Bar().init(k, x)
    y = Bar().apply(vs, x)

  def test_compact_aliasing_collision(self):
    class Foo(nn.Module):
      m1: nn.Module
      m2: nn.Module

      @nn.compact
      def __call__(self, x):
        x = self.m2(self.m1(x))
        return x

    class Bar(nn.Module):

      @nn.compact
      def __call__(self, x):
        dense = nn.Dense(2)
        x = nn.jit(Foo)(dense, dense)(x)
        return x

    k = random.key(0)
    x = jnp.zeros((2, 2))
    _ = Bar().init(k, x)

  def test_compact_aliasing_collision_arg_and_attrib(self):
    class Foo(nn.Module):
      m1: nn.Module

      @nn.compact
      def __call__(self, x, m2):
        x = m2(self.m1(x))
        return x

    class Bar(nn.Module):

      @nn.compact
      def __call__(self, x):
        dense = nn.Dense(2)
        x = nn.jit(Foo)(dense)(x, dense)
        return x

    k = random.key(0)
    x = jnp.zeros((2, 2))
    _ = Bar().init(k, x)

  def test_jit_with_setup_helpers(self):
    class Foo(nn.Module):

      def setup(self):
        self.a = nn.Dense(2)
        self.setup_helper()

      def setup_helper(self):
        self.b = nn.Dense(2)

      def __call__(self, x):
        return self.b(self.a(x))

    class JitFoo(nn.Module):

      def setup(self):
        self.a = nn.Dense(2)
        self.setup_helper()

      def setup_helper(self):
        self.b = nn.Dense(2)

      @nn.jit
      def __call__(self, x):
        return self.b(self.a(x))

    k = random.key(0)
    x = jnp.ones((2, 2))
    vs = JitFoo().init(k, x)
    y0 = JitFoo().apply(vs, x)
    vs = Foo().init(k, x)
    y1 = Foo().apply(vs, x)
    np.testing.assert_array_equal(y0, y1)

  def test_jit_kwargs(self):
    class Foo(nn.Module):

      @nn.jit
      def __call__(self, a: jax.Array, b: jax.Array):
        return a + b

    m = Foo()
    y = m.apply({}, jnp.array(1.0), b=jnp.array(2.0))

    np.testing.assert_array_equal(y, jnp.array(3.0))

  def test_jit_static_argnames(self):
    s = None

    class Foo(nn.Module):

      @partial(nn.jit, static_argnames=['b'])
      def __call__(self, a: jax.Array, b: str):
        nonlocal s
        s = b
        return a

    m = Foo()
    y = m.apply({}, jnp.array(1.0), b='hi')

    self.assertEqual(s, 'hi')
    np.testing.assert_array_equal(y, jnp.array(1.0))

  def test_jit_repr_hash(self):
    n = 0

    @partial(jax.jit, static_argnums=0)
    def f(obj):
      nonlocal n
      n += 1
      return None

    f(_HashableProxy.from_module(nn.Dense(10)))
    self.assertEqual(n, 1)
    f(_HashableProxy.from_module(nn.Dense(10)))
    self.assertEqual(n, 1)

    f(_HashableProxy.from_module(nn.Dense(20)))
    self.assertEqual(n, 2)
    f(_HashableProxy.from_module(nn.Dense(20)))
    self.assertEqual(n, 2)

  def test_jit_reuse(self):
    n = 0

    class Foo(nn.Module):

      @nn.jit
      def __call__(self, x):
        nonlocal n
        n += 1
        return x

    x = jnp.array(1.0)
    m = Foo()

    self.assertEqual(n, 0)

    y = m.apply({}, x)
    self.assertEqual(n, 1)
    y = m.apply({}, x)
    self.assertEqual(n, 1)

  @parameterized.named_parameters(('class', True), ('method', False))
  def test_jit_reuse_hash(self, jit_class: bool):
    n = 0

    class Foo(nn.Module):
      key: int

      def __call__(self, x):
        nonlocal n
        n += 1
        return x

    if jit_class:
      Foo = nn.jit(Foo)
    else:  # jit method
      Foo.__call__ = nn.jit(Foo.__call__)

    x = jnp.array(1.0)
    self.assertEqual(n, 0)

    y = Foo(1).apply({}, x)
    self.assertEqual(n, 1)
    y = Foo(1).apply({}, x)
    self.assertEqual(n, 1)
    y = Foo(2).apply({}, x)
    self.assertEqual(n, 2)
    y = Foo(2).apply({}, x)
    self.assertEqual(n, 2)

  @parameterized.named_parameters(('class', True), ('method', False))
  def test_jit_reuse_submodules(self, jit_class: bool):
    test = self
    n = 0
    key = None
    name = None

    class Foo(nn.Module):
      key: int

      def __call__(self, x):
        nonlocal n, key, name
        n += 1
        key = self.key
        name = self.name
        return x

    if jit_class:
      Foo = nn.jit(Foo)
    else:  # jit method
      Foo.__call__ = nn.jit(Foo.__call__)

    class Parent(nn.Module):

      @nn.compact
      def __call__(self, x):

        for i in range(3):
          m = Foo(i)
          y = m(x)
          test.assertEqual(key, i)
          test.assertEndsWith(name, f'Foo_{i}')
          test.assertEqual(n, i + 1)

    x = jnp.array(1.0)
    self.assertEqual(n, 0)

    y = Parent().apply({}, x)

  @parameterized.named_parameters(('class', True), ('method', False))
  def test_jit_stateful_submodules(self, jit_class: bool):
    n = 0

    class Foo(nn.Module):
      key: int

      @nn.compact
      def __call__(self, x):
        nonlocal n
        n += 1
        count = self.variable('counts', 'count', lambda: 0)
        if not self.is_initializing():
          count.value += 1
        return x

    if jit_class:
      Foo = nn.jit(Foo)
    else:  # jit method
      Foo.__call__ = nn.jit(Foo.__call__)

    class Parent(nn.Module):
      @nn.compact
      def __call__(self, x):
        for _ in range(3):
          m = Foo(0)
          x = m(x)
        return x

    m = Parent()
    x = jnp.array(1.0)
    counts = m.init({}, x)['counts']
    self.assertEqual(n, 1)

    y, updates = m.apply({'counts': counts}, x, mutable=['counts'])
    counts = updates['counts']
    self.assertEqual(n, 2)

    for count in jax.tree.leaves(counts):
      self.assertEqual(count, 1)

    y, updates = m.apply({'counts': counts}, x, mutable=['counts'])
    counts = updates['counts']
    self.assertEqual(n, 2)

    for count in jax.tree.leaves(counts):
      self.assertEqual(count, 2)

  @parameterized.named_parameters(('class', True), ('method', False))
  def test_jit_reuse_nested_submodules(self, jit_class: bool):
    test = self
    n = 0

    class Foo(nn.Module):
      key: int

      def __call__(self, x):
        return x

    class Parent(nn.Module):
      submodules: list[Foo]

      @nn.compact
      def __call__(self, x):
        nonlocal n
        n += 1
        for i, m in enumerate(self.submodules):
          x = m(x)

        return x

    if jit_class:
      Parent = nn.jit(Parent)
    else:  # jit method
      Parent.__call__ = nn.jit(Parent.__call__)

    x = jnp.array(1.0)
    self.assertEqual(n, 0)

    y = Parent([Foo(1), Foo(2)]).apply({}, x)
    self.assertEqual(n, 1)
    y = Parent([Foo(1), Foo(2)]).apply({}, x)
    self.assertEqual(n, 1)
    y = Parent([Foo(3), Foo(4)]).apply({}, x)
    self.assertEqual(n, 2)
    y = Parent([Foo(3), Foo(4)]).apply({}, x)
    self.assertEqual(n, 2)

  def test_jit_hashes_serializable_types(self):
    class Node:

      def __init__(self, a: int):
        self.a = a

      def __hash__(self):
        # test object is not being passed as static
        raise Exception('this should not be called')

      def __eq__(self, __value, /):
        raise Exception('this should not be called')

    def to_dict(node: Node):
      return {'a': node.a}

    def from_dict(node: Node, d: dict[str, Any]):
      node.a = d['a']
      return node

    serialization.register_serialization_state(Node, to_dict, from_dict)

    try:
      n = 0

      class Foo(nn.Module):
        node: Node

        @nn.jit
        @nn.compact
        def __call__(self, x):
          nonlocal n
          n += 1
          return self.node.a + nn.Dense(2)(x)

      m = Foo(Node(1))
      m.init_with_output(random.key(0), jnp.ones((2, 2)))
      self.assertEqual(n, 1)
      m.init_with_output(random.key(0), jnp.ones((2, 2)))
      self.assertEqual(n, 1)
    finally:
      del serialization._STATE_DICT_REGISTRY[Node]

  def test_while_loop(self):
    class Foo(nn.Module):

      @nn.compact
      def __call__(self, x):
        key_zero = random.key(0)
        key_zero = jnp.broadcast_to(key_zero, (2, *key_zero.shape))
        self.param('inc', lambda _: 1)
        self.put_variable('state', 'acc', 0)
        self.put_variable('state', 'rng_params', key_zero)
        self.put_variable('state', 'rng_loop', key_zero)

        def cond_fn(mdl, c):
          acc = mdl.get_variable('state', 'acc')
          return acc < x

        def body_fn(mdl, c):
          i = mdl.get_variable('state', 'acc')
          p_rng = mdl.make_rng('params')
          l_rng = mdl.make_rng('loop')
          mdl.put_variable(
              'state',
              'rng_params',
              mdl.get_variable('state', 'rng_params').at[i].set(p_rng),
          )
          mdl.put_variable(
              'state',
              'rng_loop',
              mdl.get_variable('state', 'rng_loop').at[i].set(l_rng),
          )
          inc = mdl.get_variable('params', 'inc')
          mdl.put_variable('state', 'acc', i + inc)
          return c

        return nn.while_loop(
            cond_fn,
            body_fn,
            self,
            (),
            carry_variables='state',
            split_rngs={'params': False, 'loop': True},
        )

    x = 2
    mdl = Foo()
    _, vars = mdl.apply(
        {},
        x,
        mutable=True,
        rngs={'params': random.key(1), 'loop': random.key(2)},
    )
    self.assertEqual(vars['state']['acc'], x)
    np.testing.assert_array_equal(
        vars['state']['rng_params'][0], vars['state']['rng_params'][1]
    )
    with jax_debug_key_reuse(False):
      np.testing.assert_array_compare(
          operator.__ne__,
          vars['state']['rng_loop'][0],
          vars['state']['rng_loop'][1],
      )

  def test_cond(self):
    class Foo(nn.Module):

      @nn.compact
      def __call__(self, x, pred):
        self.variable('state', 'true_count', lambda: 0)
        self.variable('state', 'false_count', lambda: 0)

        def true_fn(mdl, x):
          mdl.variable('state', 'true_count').value += 1
          return nn.Dense(2, name='dense')(x)

        def false_fn(mdl, x):
          mdl.variable('state', 'false_count').value += 1
          return -nn.Dense(2, name='dense')(x)

        return nn.cond(pred, true_fn, false_fn, self, x)

  def test_switch(self):
    class Foo(nn.Module):

      @nn.compact
      def __call__(self, x, pred):
        self.variable('state', 'a_count', lambda: 0)
        self.variable('state', 'b_count', lambda: 0)
        self.variable('state', 'c_count', lambda: 0)

        def a_fn(mdl, x):
          mdl.variable('state', 'a_count').value += 1
          return nn.Dense(2, name='dense')(x)

        def b_fn(mdl, x):
          mdl.variable('state', 'b_count').value += 1
          return -nn.Dense(2, name='dense')(x)

        def c_fn(mdl, x):
          mdl.variable('state', 'c_count').value += 1
          return nn.Dense(2, name='dense')(x)

        return nn.switch(pred, [a_fn, b_fn, c_fn], self, x)

    x = jnp.ones((1, 3))
    foo = Foo()
    y1, vars = foo.init_with_output(random.key(0), x, 0)
    self.assertEqual(vars['state'], {'a_count': 1, 'b_count': 0, 'c_count': 0})
    y2, updates = foo.apply(vars, x, 1, mutable='state')
    vars = copy(vars, updates)
    self.assertEqual(vars['state'], {'a_count': 1, 'b_count': 1, 'c_count': 0})
    np.testing.assert_allclose(y1, -y2)
    y3, updates = foo.apply(vars, x, 2, mutable='state')
    vars = copy(vars, updates)
    self.assertEqual(vars['state'], {'a_count': 1, 'b_count': 1, 'c_count': 1})
    np.testing.assert_allclose(y1, y3)

  def test_switch_multihead(self):
    class Foo(nn.Module):

      def setup(self) -> None:
        self.heads = [
            nn.Sequential([nn.Dense(10), nn.Dense(7), nn.Dense(5)]),
            nn.Sequential([nn.Dense(11), nn.Dense(5)]),
            nn.Dense(5),
        ]

      @nn.compact
      def __call__(self, x, index):
        def head_fn(i):
          def fn(mdl, x):
            mdl.variable('state', f'{i}_count', lambda: -1).value += 1
            return mdl.heads[i](x)

          return fn

        branches = [head_fn(i) for i in range(len(self.heads))]

        if self.is_mutable_collection('params'):
          for branch in branches:
            _ = branch(self, x)

        return nn.switch(index, branches, self, x)

    x = jnp.ones((1, 3))
    foo = Foo()
    y1, vars = foo.init_with_output(random.key(0), x, 0)
    self.assertEqual(vars['state'], {'0_count': 1, '1_count': 0, '2_count': 0})
    y2, updates = foo.apply(vars, x, 1, mutable='state')
    vars = copy(vars, updates)
    self.assertEqual(vars['state'], {'0_count': 1, '1_count': 1, '2_count': 0})
    y3, updates = foo.apply(vars, x, 2, mutable='state')
    vars = copy(vars, updates)
    self.assertEqual(vars['state'], {'0_count': 1, '1_count': 1, '2_count': 1})

    self.assertEqual(
        vars['params']['heads_0']['layers_0']['kernel'].shape, (3, 10)
    )
    self.assertEqual(vars['params']['heads_0']['layers_0']['bias'].shape, (10,))
    self.assertEqual(
        vars['params']['heads_0']['layers_1']['kernel'].shape, (10, 7)
    )
    self.assertEqual(vars['params']['heads_0']['layers_1']['bias'].shape, (7,))
    self.assertEqual(
        vars['params']['heads_0']['layers_2']['kernel'].shape, (7, 5)
    )
    self.assertEqual(vars['params']['heads_0']['layers_2']['bias'].shape, (5,))

    self.assertEqual(
        vars['params']['heads_1']['layers_0']['kernel'].shape, (3, 11)
    )
    self.assertEqual(vars['params']['heads_1']['layers_0']['bias'].shape, (11,))
    self.assertEqual(
        vars['params']['heads_1']['layers_1']['kernel'].shape, (11, 5)
    )
    self.assertEqual(vars['params']['heads_1']['layers_1']['bias'].shape, (5,))

    self.assertEqual(vars['params']['heads_2']['kernel'].shape, (3, 5))
    self.assertEqual(vars['params']['heads_2']['bias'].shape, (5,))

  def test_lift_instance_error(self):
    class Foo(nn.Module):

      @nn.compact
      def __call__(self, x):
        return nn.checkpoint(nn.Dense(2))(x)

    with self.assertRaises(errors.TransformTargetError):
      Foo().init(random.key(0), jnp.zeros((2, 3)))

  def test_scan_compact_count(self):
    class Foo(nn.Module):
      num_layers: int = 5

      @nn.compact
      def __call__(self, x):
        def body_fn(mdl, x):
          return nn.Dense(features=x.shape[-1])(x), ()

        x, _ = nn.scan(
            body_fn,
            length=self.num_layers,
            variable_axes={'params': 0},
            split_rngs={'params': True},
        )(self, x)
        return x

    m = Foo()
    x = jnp.ones((3,))
    v = m.init(jax.random.key(0), x)
    self.assertEqual(v['params']['Dense_0']['kernel'].shape, (5, 3, 3))
    m.apply(v, x)

  def test_bound_methods_in_direct_transforms(self):
    class CondModel(nn.Module):

      def setup(self):
        self.dense = nn.Dense(3)

      def f1(self, arr):
        arr = self.dense(arr)
        return arr

      def f2(self, arr):
        _ = self.dense(arr)
        return arr

      def __call__(self, x):
        return nn.cond(x.sum() > 0, self.f1, self.f2, self, x)

    cond_model = CondModel()

    output, init_params = jax.jit(cond_model.init_with_output)(
        jax.random.key(0), x=jnp.ones(3)
    )

  def test_add_metadata_axis(self):
    vars_copy = None

    class Foo(nn.Module):

      @nn.compact
      def __call__(self, x):
        nonlocal vars_copy
        kernel_init = nn.with_partitioning(
            nn.initializers.lecun_normal(), ('foo', 'bar')
        )
        vars_copy = self.variables
        return nn.Dense(
            4, kernel_init=kernel_init, use_bias=False, name='dense'
        )(x)

    class Test(nn.Module):

      @partial(
          nn.add_metadata_axis,
          variable_axes={'params': 0},
          metadata_params={nn.PARTITION_NAME: 'baz'},
      )
      @nn.compact
      def __call__(self, x):
        return Foo(name='foo')(x)

    k = random.key(0)
    x = jnp.ones((4, 4), dtype=jnp.float32)
    vs = Test().init(k, x)
    y = Test().apply(vs, x)
    outer_expect = jax.tree_util.tree_map(
        jnp.shape,
        freeze({
            'params': {
                'foo': {
                    'dense': {
                        'kernel': nn.Partitioned(
                            jnp.ones((4, 4)), names=('baz', 'foo', 'bar')
                        )
                    }
                }
            }
        }),
    )
    inner_expect = jax.tree_util.tree_map(
        jnp.shape,
        freeze({
            'params': {
                'dense': {
                    'kernel': nn.Partitioned(
                        jnp.ones((4, 4)), names=('foo', 'bar')
                    )
                }
            }
        }),
    )
    self.assertEqual(jax.tree_util.tree_map(jnp.shape, vs), outer_expect)
    self.assertEqual(jax.tree_util.tree_map(jnp.shape, vars_copy), inner_expect)

  def test_outer_setup_called_with_sharing_across_transforms(self):
    class A(nn.Module):

      def setup(self):
        self.foo = self.param('foo', nn.initializers.zeros, (2, 2), jnp.float32)

      def __call__(self, x):
        return self.foo

    class B(nn.Module):
      a: Any

      @nn.compact
      def __call__(self, x):
        return self.a(x)

    class C(nn.Module):

      def setup(self):
        self.a = A()
        self.b = nn.jit(B)(self.a)

      def __call__(self, x):
        b = self.b(x)
        a = self.a(x)
        return a + b

    k = random.key(0)
    x = random.randint(k, (2, 2), minval=0, maxval=10)
    vs = C().init(k, x)
    y = C().apply(vs, x)
    outer_expect = jax.tree_util.tree_map(
        jnp.shape, freeze({'params': {'a': {'foo': jnp.zeros((2, 2))}}})
    )
    self.assertEqual(jax.tree_util.tree_map(jnp.shape, vs), outer_expect)

  def test_grad_simple(self):
    class LearnScale(nn.Module):

      @nn.compact
      def __call__(self, x, y):
        p = self.param('scale', nn.initializers.ones_init(), ())
        return jnp.sum(p * x * y)

    class Foo(nn.Module):

      @nn.compact
      def __call__(self, x, y):
        x_grad, y_grad = nn.grad(
            lambda mdl, x, y: mdl(x, y), LearnScale(), x, y
        )
        return x_grad, y_grad

    x = random.uniform(random.key(1), (4,))
    y = random.uniform(random.key(2), (4,))
    vs = Foo().init(random.key(0), x, y)

    x_grad, y_grad = Foo().apply(vs, x, y)
    self.assertTrue(tree_allclose(x_grad, y))
    self.assertTrue(tree_allclose(y_grad, x))

  def test_grad_simple_with_aux(self):
    class LearnScale(nn.Module):

      @nn.compact
      def __call__(self, x, y):
        p = self.param('scale', nn.initializers.ones_init(), ())
        return jnp.sum(p * x * y), p

    class Foo(nn.Module):

      @nn.compact
      def __call__(self, x, y):
        (x_grad, y_grad), aux = nn.grad(
            lambda mdl, x, y: mdl(x, y), LearnScale(), x, y, has_aux=True
        )
        return aux, x_grad, y_grad

    x = random.uniform(random.key(1), (4,))
    y = random.uniform(random.key(2), (4,))
    vs = Foo().init(random.key(0), x, y)

    aux, x_grad, y_grad = Foo().apply(vs, x, y)
    self.assertTrue(tree_allclose(x_grad, y))
    self.assertTrue(tree_allclose(y_grad, x))
    self.assertTrue(tree_allclose(aux, vs['params']['LearnScale_0']['scale']))

  def test_value_and_grad_simple(self):
    class LearnScale(nn.Module):

      @nn.compact
      def __call__(self, x, y):
        p = self.param('scale', nn.initializers.ones_init(), ())
        return jnp.sum(p * x * y)

    class Foo(nn.Module):

      @nn.compact
      def __call__(self, x, y):
        z, (x_grad, y_grad) = nn.value_and_grad(
            lambda mdl, x, y: mdl(x, y), LearnScale(), x, y
        )
        return z, x_grad, y_grad

    x = random.uniform(random.key(1), (4,))
    y = random.uniform(random.key(2), (4,))
    vs = Foo().init(random.key(0), x, y)

    z, x_grad, y_grad = Foo().apply(vs, x, y)
    self.assertTrue(tree_allclose(x_grad, y))
    self.assertTrue(tree_allclose(y_grad, x))

  def test_value_and_grad_simple_with_aux(self):
    class LearnScale(nn.Module):

      @nn.compact
      def __call__(self, x, y):
        p = self.param('scale', nn.initializers.ones_init(), ())
        return jnp.sum(p * x * y), p

    class Foo(nn.Module):

      @nn.compact
      def __call__(self, x, y):
        (z, aux), (x_grad, y_grad) = nn.value_and_grad(
            lambda mdl, x, y: mdl(x, y), LearnScale(), x, y, has_aux=True
        )
        return z, aux, x_grad, y_grad

    x = random.uniform(random.key(1), (4,))
    y = random.uniform(random.key(2), (4,))
    vs = Foo().init(random.key(0), x, y)

    z, aux, x_grad, y_grad = Foo().apply(vs, x, y)
    self.assertTrue(tree_allclose(x_grad, y))
    self.assertTrue(tree_allclose(y_grad, x))
    self.assertTrue(tree_allclose(aux, vs['params']['LearnScale_0']['scale']))

  def test_value_and_grad_multiscope(self):
    class Foo(nn.Module):
      bar: nn.Module

      @nn.compact
      def __call__(self, x, y):
        def fn(self, x, y):
          qup = nn.Dense(y.shape[-1])
          delta = y - self.bar(qup(x))
          return jnp.sum(delta**2)

        z, (x_grad, y_grad) = nn.value_and_grad(fn, self, x, y)
        return z, x_grad, y_grad

    class Baz(nn.Module):

      @nn.compact
      def __call__(self, x, y):
        bar = nn.Dense(y.shape[-1])
        return Foo(bar=bar)(x, y)

    x = random.uniform(random.key(1), (4,))
    y = random.uniform(random.key(2), (4,))
    vs = Baz().init(random.key(0), x, y)
    z, x_grad, y_grad = Baz().apply(vs, x, y)

    def comparison_fn(x, y):
      w1 = vs['params']['Foo_0']['Dense_0']['kernel']
      w2 = vs['params']['Dense_0']['kernel']
      delta = y - jnp.dot(jnp.dot(x, w1), w2)
      return jnp.sum(delta**2)

    self.assertTrue(tree_allclose(comparison_fn(x, y), z))
    self.assertTrue(tree_allclose(jax.grad(comparison_fn, 0)(x, y), x_grad))
    self.assertTrue(tree_allclose(jax.grad(comparison_fn, 1)(x, y), y_grad))

  def test_value_and_grad_multiscope_adopted(self):
    class Foo(nn.Module):
      bar: nn.Module
      qup: nn.Module

      @nn.compact
      def __call__(self, x, y):
        def fn(self, x, y):
          delta = y - self.bar(self.qup(x))
          return jnp.sum(delta**2)

        z, (x_grad, y_grad) = nn.value_and_grad(fn, self, x, y)
        return z, x_grad, y_grad

    x = random.uniform(random.key(1), (4,))
    y = random.uniform(random.key(2), (4,))
    vs = Foo(bar=nn.Dense(4), qup=nn.Dense(4)).init(random.key(0), x, y)
    z, x_grad, y_grad = Foo(bar=nn.Dense(4), qup=nn.Dense(4)).apply(vs, x, y)

    def comparison_fn(x, y):
      w1 = vs['params']['qup']['kernel']
      w2 = vs['params']['bar']['kernel']
      delta = y - jnp.dot(jnp.dot(x, w1), w2)
      return jnp.sum(delta**2)

    self.assertTrue(tree_allclose(comparison_fn(x, y), z))
    self.assertTrue(tree_allclose(jax.grad(comparison_fn, 0)(x, y), x_grad))
    self.assertTrue(tree_allclose(jax.grad(comparison_fn, 1)(x, y), y_grad))

  def test_vmap_add_remove_axis_transforms(self):
    class BoxedData(struct.PyTreeNode, AxisMetadata):
      value: Any
      def unbox(self):
        return self.value
      def replace_boxed(self, val):
        return self.replace(value=val)
      def add_axis(self, index: int, params: Dict[Any, Any]):
        value = jnp.mean(self.value, axis=index)
        return self.replace(value=value)
      def remove_axis(self, index: int, params: Dict[Any, Any]):
        value_shape = list(self.value.shape)
        value_shape.insert(index, params['axis_size'])
        value = jnp.broadcast_to(self.value, value_shape)
        return self.replace(value=value)

    class Top(nn.Module):
      @nn.compact
      def __call__(self, x):
        VFoo = nn.vmap(
            Foo, in_axes=0, out_axes=0, variable_axes={'params':0, 'aux': 0},
            metadata_params={'axis_size': x.shape[0]},
        )
        vfoo = VFoo(name="vfoo")
        y = vfoo(x)
        y = vfoo(x)
        assert vfoo.variables['aux']['v'].value.shape == ()
        return y

    class Foo(nn.Module):
      @nn.compact
      def __call__(self, x):
        if self.has_variable('aux', 'v'):
          assert self.variables['aux']['v'].value.shape == ()
        boxed_v = self.variable('aux', 'v', lambda: BoxedData(jnp.ones(())))
        assert self.variables['aux']['v'].value.shape == ()
        return x

    vs = Top().init(random.key(0), jnp.ones((2,5)))
    y = Top().apply(vs, jnp.ones((2, 5)))
    assert vs['aux']['vfoo']['v'].value.shape == ()


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Flax implementation of ResNet V1.5."""

# See issue #620.
# pytype: disable=wrong-arg-count

from functools import partial
from typing import Any, Callable, Sequence, Tuple

from flax import linen as nn
import jax.numpy as jnp

ModuleDef = Any


class ResNetBlock(nn.Module):
  """ResNet block."""

  filters: int
  conv: ModuleDef
  norm: ModuleDef
  act: Callable
  strides: Tuple[int, int] = (1, 1)

  @nn.compact
  def __call__(
      self,
      x,
  ):
    residual = x
    y = self.conv(self.filters, (3, 3), self.strides)(x)
    y = self.norm()(y)
    y = self.act(y)
    y = self.conv(self.filters, (3, 3))(y)
    y = self.norm(scale_init=nn.initializers.zeros_init())(y)

    if residual.shape != y.shape:
      residual = self.conv(
          self.filters, (1, 1), self.strides, name='conv_proj'
      )(residual)
      residual = self.norm(name='norm_proj')(residual)

    return self.act(residual + y)


class BottleneckResNetBlock(nn.Module):
  """Bottleneck ResNet block."""

  filters: int
  conv: ModuleDef
  norm: ModuleDef
  act: Callable
  strides: Tuple[int, int] = (1, 1)

  @nn.compact
  def __call__(self, x):
    residual = x
    y = self.conv(self.filters, (1, 1))(x)
    y = self.norm()(y)
    y = self.act(y)
    y = self.conv(self.filters, (3, 3), self.strides)(y)
    y = self.norm()(y)
    y = self.act(y)
    y = self.conv(self.filters * 4, (1, 1))(y)
    y = self.norm(scale_init=nn.initializers.zeros_init())(y)

    if residual.shape != y.shape:
      residual = self.conv(
          self.filters * 4, (1, 1), self.strides, name='conv_proj'
      )(residual)
      residual = self.norm(name='norm_proj')(residual)

    return self.act(residual + y)


class ResNet(nn.Module):
  """ResNetV1.5."""

  stage_sizes: Sequence[int]
  block_cls: ModuleDef
  num_classes: int
  num_filters: int = 64
  dtype: Any = jnp.float32
  act: Callable = nn.relu
  conv: ModuleDef = nn.Conv

  @nn.compact
  def __call__(self, x, train: bool = True):
    conv = partial(self.conv, use_bias=False, dtype=self.dtype)
    norm = partial(
        nn.BatchNorm,
        use_running_average=not train,
        momentum=0.9,
        epsilon=1e-5,
        dtype=self.dtype,
        axis_name='batch',
    )

    x = conv(
        self.num_filters,
        (7, 7),
        (2, 2),
        padding=[(3, 3), (3, 3)],
        name='conv_init',
    )(x)
    x = norm(name='bn_init')(x)
    x = nn.relu(x)
    x = nn.max_pool(x, (3, 3), strides=(2, 2), padding='SAME')
    for i, block_size in enumerate(self.stage_sizes):
      for j in range(block_size):
        strides = (2, 2) if i > 0 and j == 0 else (1, 1)
        x = self.block_cls(
            self.num_filters * 2**i,
            strides=strides,
            conv=conv,
            norm=norm,
            act=self.act,
        )(x)
    x = jnp.mean(x, axis=(1, 2))
    x = nn.Dense(self.num_classes, dtype=self.dtype)(x)
    x = jnp.asarray(x, self.dtype)
    return x


ResNet18 = partial(ResNet, stage_sizes=[2, 2, 2, 2], block_cls=ResNetBlock)
ResNet34 = partial(ResNet, stage_sizes=[3, 4, 6, 3], block_cls=ResNetBlock)
ResNet50 = partial(
    ResNet, stage_sizes=[3, 4, 6, 3], block_cls=BottleneckResNetBlock
)
ResNet101 = partial(
    ResNet, stage_sizes=[3, 4, 23, 3], block_cls=BottleneckResNetBlock
)
ResNet152 = partial(
    ResNet, stage_sizes=[3, 8, 36, 3], block_cls=BottleneckResNetBlock
)
ResNet200 = partial(
    ResNet, stage_sizes=[3, 24, 36, 3], block_cls=BottleneckResNetBlock
)


ResNet18Local = partial(
    ResNet, stage_sizes=[2, 2, 2, 2], block_cls=ResNetBlock, conv=nn.ConvLocal
)


# Used for testing only.
_ResNet1 = partial(ResNet, stage_sizes=[1], block_cls=ResNetBlock)
_ResNet1Local = partial(
    ResNet, stage_sizes=[1], block_cls=ResNetBlock, conv=nn.ConvLocal
)



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Copyright 2023 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import typing as tp

from flax import struct
from flax.nnx.nnx import filterlib, graph
from flax.nnx.nnx.object import Object
from flax.nnx.nnx.variables import Variable
import jax, jax.numpy as jnp

# TODO: add tests and docstrings


class MetricState(Variable):
  """Wrapper class for Metric Variables."""

  pass


class Metric(Object):
  """Base class for metrics. Any class that subclasses ``Metric`` should
  implement a ``compute``, ``reset`` and ``update`` method."""

  def __init__(self):
    raise NotImplementedError('Must override `__init__()` method.')

  def reset(self) -> None:
    """In-place reset the ``Metric``."""
    raise NotImplementedError('Must override `reset()` method.')

  def update(self, **kwargs) -> None:
    """In-place update the ``Metric``."""
    raise NotImplementedError('Must override `update()` method.')

  def compute(self):
    """Compute and return the value of the ``Metric``."""
    raise NotImplementedError('Must override `compute()` method.')

  def split(self, *filters: filterlib.Filter):
    return graph.split(self, *filters)


class Average(Metric):
  """Average metric.

  Example usage::

    >>> import jax.numpy as jnp
    >>> from flax import nnx

    >>> batch_loss = jnp.array([1, 2, 3, 4])
    >>> batch_loss2 = jnp.array([3, 2, 1, 0])

    >>> metrics = nnx.metrics.Average()
    >>> metrics.compute()
    Array(nan, dtype=float32)
    >>> metrics.update(values=batch_loss)
    >>> metrics.compute()
    Array(2.5, dtype=float32)
    >>> metrics.update(values=batch_loss2)
    >>> metrics.compute()
    Array(2., dtype=float32)
    >>> metrics.reset()
    >>> metrics.compute()
    Array(nan, dtype=float32)
  """

  def __init__(self, argname: str = 'values'):
    """Pass in a string denoting the key-word argument that :func:`update` will use to derive the new value.
    For example, constructing the metric as ``avg = Average('test')`` would allow you to make updates with
    ``avg.update(test=new_value)``.

    Args:
      argname: an optional string denoting the key-word argument that
        :func:`update` will use to derive the new value. Defaults to
        ``'values'``.
    """
    self.argname = argname
    self.total = MetricState(jnp.array(0, dtype=jnp.float32))
    self.count = MetricState(jnp.array(0, dtype=jnp.int32))

  def reset(self) -> None:
    """Reset this ``Metric``."""
    self.total.value = jnp.array(0, dtype=jnp.float32)
    self.count.value = jnp.array(0, dtype=jnp.int32)

  def update(self, **kwargs) -> None:
    """In-place update this ``Metric``. This method will use the value from
    ``kwargs[self.argname]`` to update the metric, where ``self.argname`` is
    defined on construction.

    Args:
      **kwargs: the key-word arguments that contains a ``self.argname``
        entry that maps to the value we want to use to update this metric.
    """
    if self.argname not in kwargs:
      raise TypeError(f"Expected keyword argument '{self.argname}'")
    values: tp.Union[int, float, jax.Array] = kwargs[self.argname]
    self.total.value += (
        values if isinstance(values, (int, float)) else values.sum()
    )
    self.count.value += 1 if isinstance(values, (int, float)) else values.size

  def compute(self) -> jax.Array:
    """Compute and return the average."""
    return self.total.value / self.count.value


@struct.dataclass
class Statistics:
  mean: jnp.float32
  standard_error_of_mean: jnp.float32
  standard_deviation: jnp.float32


class Welford(Metric):
  """Uses Welford's algorithm to compute the mean and variance of a stream of data.

  Example usage::

    >>> import jax.numpy as jnp
    >>> from flax import nnx

    >>> batch_loss = jnp.array([1, 2, 3, 4])
    >>> batch_loss2 = jnp.array([3, 2, 1, 0])

    >>> metrics = nnx.metrics.Welford()
    >>> metrics.compute()
    Statistics(mean=Array(0., dtype=float32), standard_error_of_mean=Array(nan, dtype=float32), standard_deviation=Array(nan, dtype=float32))
    >>> metrics.update(values=batch_loss)
    >>> metrics.compute()
    Statistics(mean=Array(2.5, dtype=float32), standard_error_of_mean=Array(0.559017, dtype=float32), standard_deviation=Array(1.118034, dtype=float32))
    >>> metrics.update(values=batch_loss2)
    >>> metrics.compute()
    Statistics(mean=Array(2., dtype=float32), standard_error_of_mean=Array(0.43301272, dtype=float32), standard_deviation=Array(1.2247449, dtype=float32))
    >>> metrics.reset()
    >>> metrics.compute()
    Statistics(mean=Array(0., dtype=float32), standard_error_of_mean=Array(nan, dtype=float32), standard_deviation=Array(nan, dtype=float32))
  """

  def __init__(self, argname: str = 'values'):
    """Pass in a string denoting the key-word argument that :func:`update` will use to derive the new value.
    For example, constructing the metric as ``wf = Welford('test')`` would allow you to make updates with
    ``wf.update(test=new_value)``.

    Args:
      argname: an optional string denoting the key-word argument that
        :func:`update` will use to derive the new value. Defaults to
        ``'values'``.
    """
    self.argname = argname
    self.count = MetricState(jnp.array(0, dtype=jnp.int32))
    self.mean = MetricState(jnp.array(0, dtype=jnp.float32))
    self.m2 = MetricState(jnp.array(0, dtype=jnp.float32))

  def reset(self) -> None:
    """Reset this ``Metric``."""
    self.count.value = jnp.array(0, dtype=jnp.uint32)
    self.mean.value = jnp.array(0, dtype=jnp.float32)
    self.m2.value = jnp.array(0, dtype=jnp.float32)

  def update(self, **kwargs) -> None:
    """In-place update this ``Metric``. This method will use the value from
    ``kwargs[self.argname]`` to update the metric, where ``self.argname`` is
    defined on construction.

    Args:
      **kwargs: the key-word arguments that contains a ``self.argname``
        entry that maps to the value we want to use to update this metric.
    """
    if self.argname not in kwargs:
      raise TypeError(f"Expected keyword argument '{self.argname}'")
    values: tp.Union[int, float, jax.Array] = kwargs[self.argname]
    count = 1 if isinstance(values, (int, float)) else values.size
    original_count = self.count.value
    self.count.value += count
    delta = (
        values if isinstance(values, (int, float)) else values.mean()
    ) - self.mean.value
    self.mean.value += delta * count / self.count.value
    m2 = 0.0 if isinstance(values, (int, float)) else values.var() * count
    self.m2.value += (
        m2 + delta * delta * count * original_count / self.count
    )

  def compute(self) -> Statistics:
    """Compute and return the mean and variance statistics in a
    ``Statistics`` dataclass object.
    """
    variance = self.m2 / self.count
    standard_deviation = variance**0.5
    sem = standard_deviation / (self.count**0.5)
    return Statistics(
        mean=self.mean.value,
        standard_error_of_mean=sem,
        standard_deviation=standard_deviation,
    )


class Accuracy(Average):
  """Accuracy metric. This metric subclasses :class:`Average`,
  and so they share the same ``reset`` and ``compute`` method
  implementations. Unlike :class:`Average`, no string needs to
  be passed to ``Accuracy`` during construction.

  Example usage::

    >>> from flax import nnx
    >>> import jax, jax.numpy as jnp

    >>> logits = jax.random.normal(jax.random.key(0), (5, 2))
    >>> labels = jnp.array([1, 1, 0, 1, 0])
    >>> logits2 = jax.random.normal(jax.random.key(1), (5, 2))
    >>> labels2 = jnp.array([0, 1, 1, 1, 1])

    >>> metrics = nnx.metrics.Accuracy()
    >>> metrics.compute()
    Array(nan, dtype=float32)
    >>> metrics.update(logits=logits, labels=labels)
    >>> metrics.compute()
    Array(0.6, dtype=float32)
    >>> metrics.update(logits=logits2, labels=labels2)
    >>> metrics.compute()
    Array(0.7, dtype=float32)
    >>> metrics.reset()
    >>> metrics.compute()
    Array(nan, dtype=float32)
  """

  def update(self, *, logits: jax.Array, labels: jax.Array, **_) -> None:  # type: ignore[override]
    """In-place update this ``Metric``.

    Args:
      logits: the outputted predicted activations. These values are
        argmax-ed (on the trailing dimension), before comparing them
        to the labels.
      labels: the ground truth integer labels.
    """
    if logits.ndim != labels.ndim + 1 or labels.dtype != jnp.int32:
      raise ValueError(
          f'Expected labels.dtype==jnp.int32 and logits.ndim={logits.ndim}=='
          f'labels.ndim+1={labels.ndim + 1}'
      )
    super().update(values=(logits.argmax(axis=-1) == labels))


class MultiMetric(Metric):
  """MultiMetric class to store multiple metrics and update them in a single call.

  Example usage::

    >>> from flax import nnx
    >>> import jax, jax.numpy as jnp

    >>> metrics = nnx.MultiMetric(
    ...   accuracy=nnx.metrics.Accuracy(), loss=nnx.metrics.Average()
    ... )

    >>> metrics
    MultiMetric(
      accuracy=Accuracy(
        argname='values',
        total=MetricState(
          value=Array(0., dtype=float32)
        ),
        count=MetricState(
          value=Array(0, dtype=int32)
        )
      ),
      loss=Average(
        argname='values',
        total=MetricState(
          value=Array(0., dtype=float32)
        ),
        count=MetricState(
          value=Array(0, dtype=int32)
        )
      )
    )

    >>> metrics.accuracy
    Accuracy(
      argname='values',
      total=MetricState(
        value=Array(0., dtype=float32)
      ),
      count=MetricState(
        value=Array(0, dtype=int32)
      )
    )

    >>> metrics.loss
    Average(
      argname='values',
      total=MetricState(
        value=Array(0., dtype=float32)
      ),
      count=MetricState(
        value=Array(0, dtype=int32)
      )
    )

    >>> logits = jax.random.normal(jax.random.key(0), (5, 2))
    >>> labels = jnp.array([1, 1, 0, 1, 0])
    >>> logits2 = jax.random.normal(jax.random.key(1), (5, 2))
    >>> labels2 = jnp.array([0, 1, 1, 1, 1])

    >>> batch_loss = jnp.array([1, 2, 3, 4])
    >>> batch_loss2 = jnp.array([3, 2, 1, 0])

    >>> metrics.compute()
    {'accuracy': Array(nan, dtype=float32), 'loss': Array(nan, dtype=float32)}
    >>> metrics.update(logits=logits, labels=labels, values=batch_loss)
    >>> metrics.compute()
    {'accuracy': Array(0.6, dtype=float32), 'loss': Array(2.5, dtype=float32)}
    >>> metrics.update(logits=logits2, labels=labels2, values=batch_loss2)
    >>> metrics.compute()
    {'accuracy': Array(0.7, dtype=float32), 'loss': Array(2., dtype=float32)}
    >>> metrics.reset()
    >>> metrics.compute()
    {'accuracy': Array(nan, dtype=float32), 'loss': Array(nan, dtype=float32)}
  """

  def __init__(self, **metrics):
    """Pass in key-word arguments to the constructor, e.g.
    ``MultiMetric(keyword1=Average(), keyword2=Accuracy(), ...)``.

    Args:
      **metrics: the key-word arguments that will be used to access
        the corresponding ``Metric``.
    """
    # TODO: raise error if a kwarg is passed that is in ('reset', 'update', 'compute'), since these names are reserved for methods
    self._metric_names = []
    for metric_name, metric in metrics.items():
      self._metric_names.append(metric_name)
      vars(self)[metric_name] = metric

  def reset(self) -> None:
    """Reset all underlying ``Metric``'s."""
    for metric_name in self._metric_names:
      getattr(self, metric_name).reset()

  def update(self, **updates) -> None:
    """In-place update all underlying ``Metric``'s in this ``MultiMetric``. All
    ``**updates`` will be passed to the ``update`` method of all underlying
    ``Metric``'s.

    Args:
      **updates: the key-word arguments that will be passed to the underlying ``Metric``'s
        ``update`` method.
    """
    # TODO: should we give the option of updating only some of the metrics and not all? e.g. if for some kwargs==None, don't do update
    # TODO: should we raise an error if a kwarg is passed into **updates that has no match with any underlying metric? e.g. user typo
    for metric_name in self._metric_names:
      getattr(self, metric_name).update(**updates)

  def compute(self) -> dict[str, Metric]:
    """Compute and return the value of all underlying ``Metric``'s. This method
    will return a dictionary, mapping strings (defined by the key-word arguments
    ``**metrics`` passed to the constructor) to the corresponding metric value.
    """
    return {
        f'{metric_name}': getattr(self, metric_name).compute()
        for metric_name in self._metric_names
    }



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib.util

penzai_installed = importlib.util.find_spec('penzai') is not None
try:
  from IPython import get_ipython

  in_ipython = get_ipython() is not None
except ImportError:
  in_ipython = False


def display(*args):
  """Display the given objects using a Penzai visualizer.

  If Penzai is not installed or the code is not running in IPython, ``display``
  will print the objects instead.
  """
  if not penzai_installed or not in_ipython:
    for x in args:
      print(x)
    return

  from penzai import pz  # type: ignore[import-not-found,import-untyped]

  with pz.ts.active_autovisualizer.set_scoped(pz.ts.ArrayAutovisualizer()):
    for x in args:
      pz.ts.display(x, ignore_exceptions=True)
